/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu
  

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

**********************************/





#include "PLA.h"
#include <math.h>
#include <stdio.h>
#include "PLA_Blas1.h"
/******************************************************************************/

/* NOTE : Not complete -- these routines do not as yet do the sequential check*/

int PLA_Axpy_enter(PLA_Obj alpha, PLA_Obj x, PLA_Obj y)
{
	int check_parameter, instrument;
	char current_routine[20] = "PLA_Axpy";
	int objtype;
	MPI_Datatype datatype;
	PLA_Template template;
	int  error = 0, original_error = 0;

	PLA_Stack_push_current( current_routine );

	PLA_Get_check_parameters ( &check_parameter );

	if ( PLA_CHECK_PARAMETER == check_parameter )
	{
	/* check objects for validity */
		PLA_Obj_object_valid( alpha, "alpha", error, original_error);
		PLA_Obj_object_valid( x, "x", error, original_error);
		PLA_Obj_object_valid( y, "y", error, original_error);

	 /* check datatypes */
		PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
		PLA_Obj_datatype( alpha, &datatype );
		PLA_Obj_datatype_equal( x, datatype, "x", error, original_error );
		PLA_Obj_datatype_equal( y, datatype, "y", error, original_error );

	 /* Global call alpha must be a Multi-scalar */
		PLA_Obj_objtype_equal( alpha, PLA_MSCALAR, "alpha", error, original_error );

	 /* check if templates are all equal (temporary restriction) */
	 /* Copy can work with different templates 03/12/97 so commenting out JAG */
/*		PLA_Obj_template( alpha, &template );
		PLA_Obj_template_equal( x, template, "x", error, original_error );
		PLA_Obj_template_equal( y, template, "y", error, original_error ); */

	 /* check dimensions */
		PLA_Obj_global_length_equal  ( alpha, 1, "alpha", error, original_error );
		PLA_Obj_global_width_equal   ( alpha, 1, "alpha", error, original_error );

		PLA_Obj_global_length_length_equal( x, y, "x and y", error, original_error );
		PLA_Obj_global_width_width_equal  ( x, y, "x and y", error, original_error );
	}
	PLA_Get_instrument ( &instrument );
	if ( PLA_INSTRUMENT == instrument )
	{
	  /* insert instrumentation calls */
	}
	return(original_error);
}

int PLA_Axpy_exit(PLA_Obj alpha, PLA_Obj x, PLA_Obj y)
{
	int instrument;
	char current_routine[20];

	PLA_Get_instrument ( &instrument );
	if ( PLA_INSTRUMENT == instrument )
	{
	  /* insert instrumentation calls */
	}
	PLA_Stack_pop_current( current_routine );
	/* Free all of the "in" objects -- not implemented YET!! */
	return(PLA_SUCCESS);
}
/******************************************************************************/

int PLA_Axpy(PLA_Obj Alpha,
				PLA_Obj X,
				PLA_Obj Y)

{
	int error_report = 0;
	int object_type_X, object_type_Y;
	int owner_row_Alpha, owner_col_Alpha;
	int owner_row_Y, owner_col_Y;
	int owner_row_X, owner_col_X;
        int width_Y, length_Y;
        int Projection_Y;
	int row_align_X = -1, col_align_X = -1; /* Need to know if N.A. */
	int row_align_Y = -1, col_align_Y = -1;
	PLA_Obj Alpha_dup = NULL, Conformal_to_Y = NULL;
	MPI_Datatype datatype;
	PLA_Template template_Alpha, template_X, template_Y; /* Copy can work with diff templates */

	error_report += PLA_Axpy_enter(Alpha, X, Y);

/* Take care of the multiscalar duplication */

	PLA_Obj_datatype (Alpha, &datatype);
	PLA_Obj_template (Alpha, &template_Alpha);
	PLA_Obj_owner_row(Alpha, &owner_row_Alpha);
	PLA_Obj_owner_col(Alpha, &owner_col_Alpha);
	if( (owner_row_Alpha != PLA_ALL_ROWS) || (owner_col_Alpha != PLA_ALL_COLS) )
	{
		PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
								  template_Alpha,   &Alpha_dup);
		PLA_Copy(Alpha, Alpha_dup);
	}

/* Take care of local conformation */
	PLA_Obj_template( X, &template_X);
	PLA_Obj_template( Y, &template_Y);

	PLA_Obj_objtype( X, &object_type_X);
	PLA_Obj_objtype( Y, &object_type_Y);

	PLA_Obj_global_align_row(X, &row_align_X);
	PLA_Obj_global_align_col(X, &col_align_X);

	PLA_Obj_global_align_row(Y, &row_align_Y);
	PLA_Obj_global_align_col(Y, &col_align_Y);

	PLA_Obj_owner_row(X, &owner_row_X);
	PLA_Obj_owner_col(X, &owner_col_X);
	PLA_Obj_owner_row(Y, &owner_row_X);
	PLA_Obj_owner_col(Y, &owner_col_X);

	if( (template_X != template_Y) ||
		 (object_type_X != object_type_Y) ||
                 (owner_row_X != owner_row_Y) ||
                 (owner_col_X != owner_col_Y) ||
		 (row_align_X != row_align_Y)  ||
		 (col_align_X != col_align_Y) )
	{
          PLA_Obj_clone(Y, &Conformal_to_Y);
	  PLA_Copy(X, Conformal_to_Y);
	}
	PLA_Local_axpy(Alpha_dup == NULL? Alpha : Alpha_dup,
			Conformal_to_Y == NULL? X : Conformal_to_Y, Y);
	PLA_Obj_free(&Alpha_dup);
	PLA_Obj_free(&Conformal_to_Y);
	error_report += PLA_Axpy_exit (Alpha, X, Y);
	return(error_report);

}
