/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu


  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/




#include "PLA.h"
#include <math.h>
#include <stdio.h>
#include "PLA_Blas1.h"

static PLA_Obj alpha_in = NULL, x_in = NULL;

int PLA_Inv_scal_enter ( PLA_Obj alpha, PLA_Obj x )
{
   int instrument, check_parameter, check_seq, myrow, mycol;
   char current_routine[20] = "PLA_Inv_scal";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {

    /* check objects for validity */
      PLA_Obj_object_valid( alpha,  "alpha", error, original_error );
      PLA_Obj_object_valid( x,      "x", error, original_error );

    /* check object types */
      PLA_Obj_objtype_equal( alpha, PLA_MSCALAR, "alpha", error, original_error );

    /* check datatypes */
      PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
      PLA_Obj_datatype( alpha, &datatype );
      PLA_Obj_datatype_equal( x,     datatype, "x", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( alpha, &template );
      PLA_Obj_template_equal( x,     template, "x", error, original_error );

    /* check dimensions */
      PLA_Obj_global_length_equal ( alpha, 1, "alpha", error, original_error );
      PLA_Obj_global_width_equal  ( alpha, 1, "alpha", error, original_error );
   }
   PLA_Get_check_sequential ( &check_seq );
   if ( PLA_CHECK_SEQUENTIAL == check_seq )
   {
       PLA_Mscalar_create_conf_to( alpha, 0, 0, &alpha_in );
       PLA_Mscalar_create_conf_to( x,     0, 0, &x_in );

       PLA_Copy_temp( alpha, alpha_in );
       PLA_Copy_temp( x,     x_in );

       PLA_Temp_comm_col_rank( template, &myrow );
       PLA_Temp_comm_row_rank( template, &mycol );

       if ( 0 == myrow && 0 == mycol )
          PLA_Local_inv_scal( alpha_in, x_in );
   }

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
}


int PLA_Inv_scal_exit ( PLA_Obj alpha, PLA_Obj x )
{
  char current_routine[20];
  int  check_seq, instrument, myrow, mycol;
  PLA_Obj x_out = NULL;
  PLA_Template template;

  PLA_Get_check_sequential ( &check_seq );
/* MAY (JAG) want to test the objects for validity here as well      */
/* in case we have a wayward pointer or some other memory corruption */
  if ( PLA_CHECK_SEQUENTIAL == check_seq )
  {
     PLA_Mscalar_create_conf_to( x,     0, 0, &x_out );
     PLA_Copy_temp(x, x_out);
     PLA_Obj_template(x, &template);
     PLA_Temp_comm_col_rank( template, &myrow );
     PLA_Temp_comm_row_rank( template, &mycol );
     if ( 0 == myrow && 0 == mycol )
     {
/* Commented out because I know we have the code to do this somewhere */
/* and I would like to avoid re-writing it.  It's not a big deal to   */
/* rewrite if we restrict what can be passed to the function though.  */
/*        diff = PLA_max_abs_diff( x_in, x_out ); */
      }
   }

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   PLA_Obj_free( &alpha_in );
   PLA_Obj_free( &x_in  );
   PLA_Obj_free( &x_out  );
   return(PLA_SUCCESS);
}

/********************************************************************/
/* PLA_Inv_scal() takes a multi-scalar and an object and scales the */
/* object by the inverse of the multi-scalar                        */
/********************************************************************/
int PLA_Inv_scal(PLA_Obj alpha,
            PLA_Obj x)
{

   int owner_row, owner_col;
   PLA_Obj alpha_dup = NULL;


   PLA_Inv_scal_enter(alpha, x);
   PLA_Obj_owner_row(alpha, &owner_row);
   PLA_Obj_owner_col(alpha, &owner_col);
   if( (owner_row != PLA_ALL_ROWS) || (owner_col != PLA_ALL_COLS) )
   {
      PLA_Mscalar_create_conf_to(alpha, PLA_ALL_ROWS, PLA_ALL_COLS, &alpha_dup);
      PLA_Copy(alpha, alpha_dup);
   }

   PLA_Local_inv_scal( alpha_dup == NULL ? alpha : alpha_dup, x);

   PLA_Obj_free( &alpha_dup );

   PLA_Inv_scal_exit(alpha, x);
   return 0;
}
