#include "PLA.h"
#ifndef min3
#define min3(x, y, z)        min(min(x, y), z)
#endif

#ifndef PLASUM
#define PLASUM(datatype) ( datatype == MPI_DOUBLE_COMPLEX ? PLA_SUM : MPI_SUM)
#endif
/*****************************************************************************/
int PLA_Symm_right_lower_pp(int nb_alg, PLA_Obj alpha, PLA_Obj a,
PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
   int          idummy, length_a, width_a, size, last_size;
   PLA_Template template;
   MPI_Datatype datatype;
   PLA_Obj      duplicated_b_cols = NULL,        /* Is square -- just dup */
   duplicated_a_rows = NULL,        /* Needs symmetric components */
   triangular_dup_a_rows = NULL,    /* Part with structure */
   transposed_dup_a_rows = NULL,    /* Part that trans via copy */
   nontransposed_dup_a_rows = NULL, /* Simply Bcast this part   */
   a_rows = NULL,                   /* View of source */
   a_cols = NULL,                   /* Part that will be trans */
   b_cols = NULL,                   /* View of source */
   triangular_c = NULL,             /* Labelled for local kernel */
   transposed_c = NULL,             /* Not transposed itself but */
   /* updated by transposed piece*/
   nontransposed_c = NULL,
   a_cur = NULL, /* Only used for length_a/width_b and */
   b_cur = NULL, /* reading ease.                      */
   One = NULL  ;

   /* Scale all of c by beta */
   PLA_Scal( beta, c );

   PLA_Obj_template(c, &template);
   PLA_Obj_datatype(c, &datatype);
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,
                       &One );
   PLA_Obj_set_to_one(One);

   /* take views of all of both A and B */
   PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_cur );
   PLA_Obj_view( a, 0,           0,           PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_rows );
   PLA_Obj_view( a, PLA_DIM_ALL, 0,           PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_cols );
   PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &b_cur );
   PLA_Obj_view( b, PLA_DIM_ALL, 0,           PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &b_cols );

   /* annotate the views                 */
   PLA_Obj_set_orientation( a_rows, PLA_PROJ_ONTO_ROW );
   PLA_Obj_set_orientation( a_cols, PLA_PROJ_ONTO_COL );
   PLA_Obj_set_orientation( b_cols, PLA_PROJ_ONTO_COL );

   /* These two views are initialized to the same thing but they never actually view the same */
   /* thing during calculations.  The triangular part is updated before the kernel calls and  */
   /* the transposed part it updated afterward.                                               */
   PLA_Obj_view( c, PLA_DIM_ALL, 0, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &nontransposed_c );
   PLA_Obj_view( c, PLA_DIM_ALL, 0, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &triangular_c );

   PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &transposed_c );
	PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg,
					  &duplicated_a_rows );
	PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg,
					  &duplicated_b_cols );

	PLA_Obj_view(duplicated_a_rows, PLA_DIM_ALL, 0,
					 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &nontransposed_dup_a_rows);
	PLA_Obj_view(duplicated_a_rows, PLA_DIM_ALL, 0,
					 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &triangular_dup_a_rows);
	PLA_Obj_view_all(duplicated_a_rows, &transposed_dup_a_rows);


	last_size = 0; /* we will need to update the views by what happend on the last */
	/* go-around due to structural concerns.                        */
	while ( TRUE ) {
		/* Really we only need one of these -- or the dimensions would be wrong */
		PLA_Obj_split_size ( a_cur, PLA_SIDE_TOP, &length_a, &idummy );
		PLA_Obj_split_size ( a_cur, PLA_SIDE_LEFT,  &width_a, &idummy );
		if ( ( size = min3( length_a, width_a, nb_alg )) == 0 ) break;
		PLA_Obj_view_shift(a_cur,             size,
												  0,            0,
															 0);
		PLA_Obj_view_shift(b_cur,               0,
												 size,             0,
															 0);

		PLA_Obj_view_shift(a_rows,            last_size,
												  0,               size,
															 size);
		PLA_Obj_view_shift(a_cols,              size,
											last_size,             size,
															  0);

		PLA_Obj_view_shift(b_cols,               0,
											 last_size,               size,
															  0);

		PLA_Obj_view_shift(duplicated_b_cols,               0,
																	 0,            size - nb_alg,
																			 0);
		PLA_Obj_view_shift(nontransposed_dup_a_rows,           0,
																 0,                    size,
																		  size - nb_alg);
		PLA_Obj_view_shift(triangular_dup_a_rows,            0,
															 last_size,               size,
																	  size - nb_alg);
		PLA_Obj_view_shift(nontransposed_c,                   0,
																 0,               last_size,
																				0);
		PLA_Obj_view_shift(triangular_c,                      0,
																 last_size,               size,
																				0);
		PLA_Obj_view_shift(transposed_c,                      0,
																 size,               0,
																				0);

/* spread data                   */
/* This little trick is just used to avoid two PLA_Copy calls  -- copy into the     */
/* duplicated_non_transposed piece then move down -- during the copy, the triangular */
/* part and the nontransposed part overlap -- but before the kernels are called that*/
/* gets fixed.  PLA_Copy_x doesn't exist so this seemed the right thing to do.      */
/* Actually PLA_Copy_x could be just PLA_Copy followed by my pla_scal_traps in Trsm */
		PLA_Obj_view_shift(transposed_dup_a_rows,             0,
																	 size,             0,
																		size - nb_alg );
		PLA_Copy( a_cols, transposed_dup_a_rows );

		PLA_Copy( a_rows, nontransposed_dup_a_rows);
		PLA_Obj_view_shift(nontransposed_dup_a_rows,             0,
																	 0,             -size,
																					0 );
		PLA_Copy( b_cols, duplicated_b_cols );
		/* Perform local routines */
		PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS,
			 alpha, duplicated_b_cols, nontransposed_dup_a_rows,
			 One,   nontransposed_c);
		PLA_Local_symm(PLA_SIDE_RIGHT, PLA_LOW_TRIAN,
			 alpha, triangular_dup_a_rows, duplicated_b_cols,
			 One,   triangular_c);
      PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS,
          alpha, duplicated_b_cols, transposed_dup_a_rows,
          One,   transposed_c);
      last_size = size;
      PLA_Obj_view_shift(duplicated_b_cols,               0,
                                                    0,            nb_alg-size,
                                                          0);
      PLA_Obj_view_shift(nontransposed_dup_a_rows,           0,
                                                 0,               size,
                                                        nb_alg-size);
      PLA_Obj_view_shift(triangular_dup_a_rows,            0,
                                                  0,               0,
                                                     nb_alg - size);
      PLA_Obj_view_shift(transposed_dup_a_rows,             0,
                                                    0,             0,
                                                      nb_alg - size);
   }
   /* Free the views                     */
   PLA_Obj_free( &duplicated_b_cols  );
   PLA_Obj_free( &duplicated_a_rows  );
   PLA_Obj_free( &triangular_dup_a_rows  );
   PLA_Obj_free( &transposed_dup_a_rows  );
   PLA_Obj_free( &nontransposed_dup_a_rows  );
   PLA_Obj_free( &a_rows  );
   PLA_Obj_free( &a_cols  );
   PLA_Obj_free( &b_cols  );
   PLA_Obj_free( &triangular_c  );
   PLA_Obj_free( &transposed_c  );
   PLA_Obj_free( &nontransposed_c  );
   PLA_Obj_free( &a_cur  );
   PLA_Obj_free( &b_cur  );
   PLA_Obj_free( &One );
}
/*****************************************************************************/
int PLA_Symm_right_upper_pp(int nb_alg, PLA_Obj alpha, PLA_Obj a,
PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
   int          length_a, width_a, size, last_size,idummy;
   PLA_Template template;
   MPI_Datatype datatype;
   PLA_Obj      duplicated_b_cols = NULL,        /* Is square -- just dup */
   duplicated_a_rows = NULL,        /* Needs symmetric components */
   triangular_dup_a_rows = NULL,    /* Part with structure */
   transposed_dup_a_rows = NULL,    /* Part that trans via copy */
   nontransposed_dup_a_rows = NULL, /* Simply Bcast this part   */
   a_rows = NULL,                   /* View of source */
   a_cols = NULL,                   /* Part that will be trans */
   b_cols = NULL,                   /* View of source */
   triangular_c = NULL,             /* Labelled for local kernel */
   transposed_c = NULL,             /* Not transposed itself but */
   /* updated by transposed piece*/
   nontransposed_c = NULL,
   a_cur = NULL, /* Only used for length_a/width_b and */
   b_cur = NULL, /* reading ease.                      */
   One = NULL  ;

   /* Scale all of c by beta */
   PLA_Scal( beta, c );

   PLA_Obj_template(c, &template);
   PLA_Obj_datatype(c, &datatype);
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,
                       &One );
   PLA_Obj_set_to_one(One);

   /* take views of all of both A and B */
   PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_cur );
   PLA_Obj_view( a, 0,           PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_rows );
   PLA_Obj_view( a, 0,           0,           PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_cols );
   PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &b_cur );
   PLA_Obj_view( b, PLA_DIM_ALL, 0,           PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &b_cols );

   /* annotate the views                 */
   PLA_Obj_set_orientation( a_rows, PLA_PROJ_ONTO_ROW );
   PLA_Obj_set_orientation( a_cols, PLA_PROJ_ONTO_COL );
   PLA_Obj_set_orientation( b_cols, PLA_PROJ_ONTO_COL );

   /* These two views are initialized to the same thing but they never actually view the same */
   /* thing during calculations.  The triangular part is updated before the kernel calls and  */
   /* the transposed part it updated afterward.                                               */
   PLA_Obj_view( c, PLA_DIM_ALL,     0,       PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &transposed_c );
   PLA_Obj_view( c, PLA_DIM_ALL,     0, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &triangular_c );

   PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &nontransposed_c );

   PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg,
                 &duplicated_a_rows );
   PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg,
                 &duplicated_b_cols );


   PLA_Obj_view(duplicated_a_rows, PLA_DIM_ALL, 0,
                PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &transposed_dup_a_rows);
   PLA_Obj_view(duplicated_a_rows, PLA_DIM_ALL, 0,
                PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &triangular_dup_a_rows);
   PLA_Obj_view_all(duplicated_a_rows, &nontransposed_dup_a_rows);

   last_size = 0; /* we will need to update the views by what happend on the last */
   /* go-around due to structural concerns.                        */
   while ( TRUE ) {
      /* Really we only need one of these -- or the dimensions would be wrong */
      PLA_Obj_split_size ( a_cur, PLA_SIDE_TOP,  &length_a, &idummy );
      PLA_Obj_split_size ( a_cur, PLA_SIDE_LEFT, &width_a, &idummy );
      if ( ( size = min3( length_a, width_a, nb_alg )) == 0 ) break;
      PLA_Obj_view_shift(a_cur,              size,
                                       0,             0,
                                              0);
      PLA_Obj_view_shift(b_cur,               0,
                                     size,            0,
                                              0);

      PLA_Obj_view_shift(a_rows,           last_size,
                                    last_size,       0,
                                               size);
      PLA_Obj_view_shift(a_cols,                0,
                                    last_size,            size,
                                             last_size);

      PLA_Obj_view_shift(b_cols,                0,
                                      last_size,          size,
                                                0);

      PLA_Obj_view_shift(duplicated_b_cols,               0,
                                                    0,            size-nb_alg,
                                                          0);
      PLA_Obj_view_shift(transposed_dup_a_rows,              0,
                                                 0,               last_size,
                                                        size-nb_alg);
      PLA_Obj_view_shift(triangular_dup_a_rows,              0,
                                                 last_size,       size,
                                                        size-nb_alg);
      PLA_Obj_view_shift(transposed_c,                      0,
                                                 0,               last_size,
                                                            0);
      PLA_Obj_view_shift(triangular_c,                      0,
                                                 last_size,       size,
                                                            0);
      PLA_Obj_view_shift(nontransposed_c,                   0,
                                                 size,            0,
                                                            0);
/* spread data                   */
/* This little trick is just used to avoid two PLA_Copy calls  -- copy into the     */
/* duplicated_non_transposed piece then move down -- during the copy, the triangular */
/* part and the nontransposed part overlap -- but before the kernels are called that*/
/* gets fixed.  PLA_Copy_x doesn't exist so this seemed the right thing to do.      */
/* Actually PLA_Copy_x could be just PLA_Copy followed by my pla_scal_traps in Trsm */
      PLA_Obj_view_shift(nontransposed_dup_a_rows,          0,
                                                     0,             0,
                                                        size-nb_alg);
      PLA_Copy( a_rows, nontransposed_dup_a_rows );
      PLA_Obj_view_shift(nontransposed_dup_a_rows,          0,
                                                   size,             0,
                                                            0);
      PLA_Copy( a_cols, transposed_dup_a_rows);
      PLA_Copy( b_cols, duplicated_b_cols );
      /* Perform local routines */
      PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS,
          alpha, duplicated_b_cols, transposed_dup_a_rows,
          One,   transposed_c);
      PLA_Local_symm(PLA_SIDE_RIGHT, PLA_UP_TRIAN,
          alpha, triangular_dup_a_rows, duplicated_b_cols,
          One,   triangular_c);
      PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS,
          alpha, duplicated_b_cols, nontransposed_dup_a_rows,
          One,   nontransposed_c);
      last_size = size;
      PLA_Obj_view_shift(duplicated_b_cols,               0,
                                                    0,            nb_alg-size,
                                                          0);
      PLA_Obj_view_shift(transposed_dup_a_rows,              0,
                                                 0,                0,
                                                        nb_alg-size);
      PLA_Obj_view_shift(triangular_dup_a_rows,              0,
                                                 0,                0,
                                                        nb_alg-size);
      PLA_Obj_view_shift(nontransposed_dup_a_rows,          0,
                                                     0,             0,
                                                        nb_alg-size);
   }
   /* Free the views                     */
   PLA_Obj_free( &duplicated_b_cols  );
   PLA_Obj_free( &duplicated_a_rows  );
   PLA_Obj_free( &triangular_dup_a_rows  );
   PLA_Obj_free( &transposed_dup_a_rows  );
   PLA_Obj_free( &nontransposed_dup_a_rows  );
   PLA_Obj_free( &a_rows  );
   PLA_Obj_free( &a_cols  );
   PLA_Obj_free( &b_cols  );
   PLA_Obj_free( &triangular_c  );
   PLA_Obj_free( &transposed_c  );
   PLA_Obj_free( &nontransposed_c  );
   PLA_Obj_free( &a_cur  );
   PLA_Obj_free( &b_cur  );
   PLA_Obj_free( &One );
}

/*****************************************************************************/
int PLA_Symm_left_upper_pp(int nb_alg, PLA_Obj alpha, PLA_Obj a,
PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
   int          width_a, length_a, size, last_size, idummy;
   PLA_Template template;
   MPI_Datatype datatype;
   PLA_Obj      duplicated_b_rows = NULL,        /* Is square -- just dup */
   duplicated_a_cols = NULL,        /* Needs symmetric components */
   triangular_dup_a_cols = NULL,    /* Part with structure */
   transposed_dup_a_cols = NULL,    /* Part that trans via copy */
   nontransposed_dup_a_cols = NULL, /* Simply Bcast this part   */
   a_cols = NULL,                   /* View of source */
   a_rows = NULL,                   /* Part that will be trans */
   b_rows = NULL,                   /* View of source */
   triangular_c = NULL,             /* Labelled for local kernel */
   transposed_c = NULL,             /* Not transposed itself but */
   /* updated by transposed piece*/
   nontransposed_c = NULL,
   a_cur = NULL, /* Only used for width_a/length_b and */
   b_cur = NULL, /* reading ease.                      */
   One = NULL  ;

   /* Scale all of c by beta */
   PLA_Scal( beta, c );

   PLA_Obj_template(c, &template);
   PLA_Obj_datatype(c, &datatype);
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,
                       &One );
   PLA_Obj_set_to_one(One);

   /* take views of all of both A and B */
   PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_cur );
   PLA_Obj_view( a, 0,           0,           PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_cols );
   PLA_Obj_view( a, 0,       PLA_DIM_ALL,     PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_rows );
   PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &b_cur );
   PLA_Obj_view( b, 0,           PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &b_rows );

   /* annotate the views                 */
   PLA_Obj_set_orientation( a_cols, PLA_PROJ_ONTO_COL );
   PLA_Obj_set_orientation( a_rows, PLA_PROJ_ONTO_ROW );
   PLA_Obj_set_orientation( b_rows, PLA_PROJ_ONTO_ROW );

   /* These two views are initialized to the same thing but they never actually view the same */
   /* thing during calculations.  The triangular part is updated before the kernel calls and  */
   /* the transposed part it updated afterward.                                               */
   PLA_Obj_view( c, 0,           PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &nontransposed_c );
   PLA_Obj_view( c, 0,           PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &triangular_c );

   PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &transposed_c );

	PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg,
					  &duplicated_a_cols );
	PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg,
					  &duplicated_b_rows );

	PLA_Obj_view(duplicated_a_cols, 0, PLA_DIM_ALL,
					 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &nontransposed_dup_a_cols);
	PLA_Obj_view(duplicated_a_cols, 0, PLA_DIM_ALL,
					 PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &triangular_dup_a_cols);
	PLA_Obj_view_all(duplicated_a_cols, &transposed_dup_a_cols);


	last_size = 0; /* we will need to update the views by what happend on the last */
	/* go-around due to structural concerns.                        */
	while ( TRUE ) {
		/* Really we only need one of these -- or the dimensions would be wrong */
		PLA_Obj_split_size ( a_cur, PLA_SIDE_TOP, &length_a, &idummy );
		PLA_Obj_split_size ( a_cur, PLA_SIDE_LEFT,  &width_a, &idummy );
		if ( ( size = min3( width_a, length_a, nb_alg )) == 0 ) break;
		PLA_Obj_view_shift(a_cur,               0,
												size,            0,
															 0);
		PLA_Obj_view_shift(b_cur,             size,
													0,             0,
															 0);

		PLA_Obj_view_shift(a_cols,               0,
											last_size,            size,
															size);
		PLA_Obj_view_shift(a_rows,           last_size,
												 size,             0,
															size);

		PLA_Obj_view_shift(b_rows,           last_size,
												  0,               0,
															 size);

		PLA_Obj_view_shift(duplicated_b_rows,               0,
																	 0,               0,
																	  size - nb_alg);
		PLA_Obj_view_shift(nontransposed_dup_a_cols,           0,
																 0,               size - nb_alg,
																			  size);
		PLA_Obj_view_shift(triangular_dup_a_cols,          last_size,
																 0,               size - nb_alg,
																				size);
		PLA_Obj_view_shift(nontransposed_c,                      0,
																 0,               0,
																		  last_size);
		PLA_Obj_view_shift(triangular_c,                  last_size,
																 0,               0,
																			  size);
		PLA_Obj_view_shift(transposed_c,                     size,
																 0,               0,
																				0);

/* spread data                   */
/* This little trick is just used to avoid two PLA_Copy calls  -- copy into the     */
/* duplicated_non_transposed piece then move down -- during the copy, the triangular */
/* part and the nontransposed part overlap -- but before the kernels are called that*/
/* gets fixed.  PLA_Copy_x doesn't exist so this seemed the right thing to do.      */
/* Actually PLA_Copy_x could be just PLA_Copy followed by my pla_scal_traps in Trsm */
		PLA_Obj_view_shift(transposed_dup_a_cols,           size,
																	0,             size - nb_alg,
																			 0 );
		PLA_Copy( a_rows, transposed_dup_a_cols );
		PLA_Copy( a_cols, nontransposed_dup_a_cols);
		PLA_Obj_view_shift(nontransposed_dup_a_cols,           0,
																	  0,             0,
																			  -size );
		PLA_Copy( b_rows, duplicated_b_rows );
		/* Perform local routines */
		PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS,
			 alpha, nontransposed_dup_a_cols, duplicated_b_rows,
			 One,   nontransposed_c);
		PLA_Local_symm(PLA_SIDE_LEFT, PLA_UP_TRIAN,
			 alpha, triangular_dup_a_cols, duplicated_b_rows,
			 One,   triangular_c);
		PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS,
			 alpha, transposed_dup_a_cols, duplicated_b_rows,
			 One,   transposed_c);
		last_size = size;
		PLA_Obj_view_shift(duplicated_b_rows,               0,
																	 0,               0,
																	  nb_alg-size);
		PLA_Obj_view_shift(nontransposed_dup_a_cols,           0,
																 0,               nb_alg-size,
																			  size);
		PLA_Obj_view_shift(triangular_dup_a_cols,              0,
																 0,               nb_alg-size,
																				 0);
		PLA_Obj_view_shift(transposed_dup_a_cols,           0,
																	0,             nb_alg-size,
																			 0 );
	}
	/* Free the views                     */
	PLA_Obj_free( &duplicated_b_rows  );
	PLA_Obj_free( &duplicated_a_cols  );
	PLA_Obj_free( &triangular_dup_a_cols  );
	PLA_Obj_free( &transposed_dup_a_cols  );
	PLA_Obj_free( &nontransposed_dup_a_cols  );
	PLA_Obj_free( &a_cols  );
	PLA_Obj_free( &a_rows  );
	PLA_Obj_free( &b_rows  );
	PLA_Obj_free( &triangular_c  );
   PLA_Obj_free( &transposed_c  );
   PLA_Obj_free( &nontransposed_c  );
   PLA_Obj_free( &a_cur  );
   PLA_Obj_free( &b_cur  );
   PLA_Obj_free( &One );
}
/*****************************************************************************/
int PLA_Symm_left_lower_pp(int nb_alg, PLA_Obj alpha, PLA_Obj a,
PLA_Obj b, PLA_Obj beta, PLA_Obj c )
{
   int          width_a, length_a, size, last_size, idummy;
   PLA_Template template;
   MPI_Datatype datatype;
   PLA_Obj      duplicated_b_rows = NULL,        /* Is square -- just dup */
   duplicated_a_cols = NULL,        /* Needs symmetric components */
   triangular_dup_a_cols = NULL,    /* Part with structure */
   transposed_dup_a_cols = NULL,    /* Part that trans via copy */
   nontransposed_dup_a_cols = NULL, /* Simply Bcast this part   */
   a_cols = NULL,                   /* View of source */
   a_rows = NULL,                   /* Part that will be trans */
   b_rows = NULL,                   /* View of source */
   triangular_c = NULL,             /* Labelled for local kernel */
   transposed_c = NULL,             /* Not transposed itself but */
   /* updated by transposed piece*/
   nontransposed_c = NULL,
   a_cur = NULL, /* Only used for width_a/length_b and */
   b_cur = NULL, /* reading ease.                      */
   One = NULL  ;

   /* Scale all of c by beta */
   PLA_Scal( beta, c );

   PLA_Obj_template(c, &template);
   PLA_Obj_datatype(c, &datatype);
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,
                       &One );
   PLA_Obj_set_to_one(One);

   /* take views of all of both A and B */
   PLA_Obj_view( a, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_cur );
   PLA_Obj_view( a, PLA_DIM_ALL,           0, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_cols );
   PLA_Obj_view( a, 0,           0,           PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &a_rows );
   PLA_Obj_view( b, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &b_cur );
   PLA_Obj_view( b, 0,           PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                &b_rows );

   /* annotate the views                 */
   PLA_Obj_set_orientation( a_cols, PLA_PROJ_ONTO_COL );
   PLA_Obj_set_orientation( a_rows, PLA_PROJ_ONTO_ROW );
   PLA_Obj_set_orientation( b_rows, PLA_PROJ_ONTO_ROW );

   /* These two views are initialized to the same thing but they never actually view the same */
   /* thing during calculations.  The triangular part is updated before the kernel calls and  */
   /* the transposed part it updated afterward.                                               */
   PLA_Obj_view( c, 0,           PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &transposed_c );
   PLA_Obj_view( c, 0,           PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &triangular_c );

   PLA_Obj_view( c, PLA_DIM_ALL, PLA_DIM_ALL, PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
                 &nontransposed_c );

   PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb_alg,
                 &duplicated_a_cols );
   PLA_Pmvector_create_conf_to( c, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb_alg,
                 &duplicated_b_rows );


   PLA_Obj_view(duplicated_a_cols, 0, PLA_DIM_ALL,
                PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &transposed_dup_a_cols);
   PLA_Obj_view(duplicated_a_cols, 0, PLA_DIM_ALL,
                PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &triangular_dup_a_cols);
   PLA_Obj_view_all(duplicated_a_cols, &nontransposed_dup_a_cols);

   last_size = 0; /* we will need to update the views by what happend on the last */
   /* go-around due to structural concerns.                        */
   while ( TRUE ) {
      /* Really we only need one of these -- or the dimensions would be wrong */
      PLA_Obj_split_size ( a_cur, PLA_SIDE_TOP,  &length_a, &idummy );
      PLA_Obj_split_size ( a_cur, PLA_SIDE_LEFT, &width_a, &idummy );
      if ( ( size = min3( width_a, length_a, nb_alg )) == 0 ) break;
      PLA_Obj_view_shift(a_cur,               0,
                                    size,            0,
                                              0);
      PLA_Obj_view_shift(b_cur,             size,
                                       0,             0,
                                              0);

      PLA_Obj_view_shift(a_cols,           last_size,
                                    last_size,      size,
                                               0);
      PLA_Obj_view_shift(a_rows,           last_size,
                                       0,            last_size,
                                             size);

      PLA_Obj_view_shift(b_rows,           last_size,
                                      0,               0,
                                              size);

      PLA_Obj_view_shift(duplicated_b_rows,               0,
                                                    0,               0,
                                                     size - nb_alg);
      PLA_Obj_view_shift(transposed_dup_a_cols,              0,
                                                 0,               size - nb_alg,
                                                        last_size);
      PLA_Obj_view_shift(triangular_dup_a_cols,          last_size,
                                                 0,               size - nb_alg,
                                                            size);
      PLA_Obj_view_shift(transposed_c,                      0,
                                                 0,               0,
                                                        last_size);
      PLA_Obj_view_shift(triangular_c,                  last_size,
                                                 0,               0,
                                                           size);
      PLA_Obj_view_shift(nontransposed_c,                  size,
                                                 0,               0,
                                                            0);
/* spread data                   */
/* This little trick is just used to avoid two PLA_Copy calls  -- copy into the     */
/* duplicated_non_transposed piece then move down -- during the copy, the triangular */
/* part and the nontransposed part overlap -- but before the kernels are called that*/
/* gets fixed.  PLA_Copy_x doesn't exist so this seemed the right thing to do.      */
/* Actually PLA_Copy_x could be just PLA_Copy followed by my pla_scal_traps in Trsm */
      PLA_Obj_view_shift(nontransposed_dup_a_cols,          0,
                                                   0,             size - nb_alg,
                                                            0);
      PLA_Copy( a_cols, nontransposed_dup_a_cols );
      PLA_Obj_view_shift(nontransposed_dup_a_cols,         size,
                                                   0,             0,
                                                            0);
      PLA_Copy( a_rows, transposed_dup_a_cols);
      PLA_Copy( b_rows, duplicated_b_rows );
      /* Perform local routines */
      PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS,
          alpha, transposed_dup_a_cols, duplicated_b_rows,
          One,   transposed_c);
      PLA_Local_symm(PLA_SIDE_LEFT, PLA_LOW_TRIAN,
          alpha, triangular_dup_a_cols, duplicated_b_rows,
          One,   triangular_c);
      PLA_Local_gemm(PLA_NO_TRANS, PLA_NO_TRANS,
          alpha, nontransposed_dup_a_cols, duplicated_b_rows,
          One,   nontransposed_c);
      last_size = size;
      PLA_Obj_view_shift(duplicated_b_rows,               0,
                                                    0,               0,
                                                     nb_alg-size);
      PLA_Obj_view_shift(transposed_dup_a_cols,              0,
                                                 0,               nb_alg-size,
                                                             0);
      PLA_Obj_view_shift(triangular_dup_a_cols,              0,
                                                 0,               nb_alg-size,
                                                             0);
      PLA_Obj_view_shift(nontransposed_dup_a_cols,          0,
                                                   0,             nb_alg-size,
                                                            0);
   }
   /* Free the views                     */
   PLA_Obj_free( &duplicated_b_rows  );
   PLA_Obj_free( &duplicated_a_cols  );
   PLA_Obj_free( &triangular_dup_a_cols  );
   PLA_Obj_free( &transposed_dup_a_cols  );
   PLA_Obj_free( &nontransposed_dup_a_cols  );
   PLA_Obj_free( &a_cols  );
   PLA_Obj_free( &a_rows  );
   PLA_Obj_free( &b_rows  );
   PLA_Obj_free( &triangular_c  );
   PLA_Obj_free( &transposed_c  );
   PLA_Obj_free( &nontransposed_c  );
   PLA_Obj_free( &a_cur  );
   PLA_Obj_free( &b_cur  );
   PLA_Obj_free( &One );
}

/*****************************************************************************/
int PLA_Symm(int side, int uplo,  PLA_Obj alpha, PLA_Obj a, PLA_Obj b,
                                  PLA_Obj beta, PLA_Obj c )
{
   PLA_Obj alpha_dup = NULL, beta_dup = NULL;
   int nb_alg, owner_row, owner_col;
   PLA_Template template;

   PLA_Obj_template(a, &template);
   PLA_Environ_nb_alg(PLA_OP_PAN_PAN, template,  &nb_alg);

   PLA_Obj_owner_row(alpha, &owner_row);
   PLA_Obj_owner_col(alpha, &owner_col);
   if ( owner_row != PLA_ALL_ROWS || owner_col != PLA_ALL_COLS )
   {
      PLA_Mscalar_create_conf_to( alpha,PLA_ALL_ROWS, PLA_ALL_COLS,&alpha_dup);
      PLA_Copy( alpha, alpha_dup );
   }
   PLA_Obj_owner_row(beta, &owner_row);
   PLA_Obj_owner_col(beta, &owner_col);
   if ( owner_row != PLA_ALL_ROWS || owner_col != PLA_ALL_COLS )
   {
      PLA_Mscalar_create_conf_to( beta,PLA_ALL_ROWS, PLA_ALL_COLS,&beta_dup);
      PLA_Copy( beta, beta_dup );
   }

   if( PLA_SIDE_LEFT == side)
   {
      if(PLA_UPPER_TRIANGULAR == uplo)
         PLA_Symm_left_upper_pp(nb_alg,
                                alpha_dup == NULL ? alpha : alpha_dup, a, b,
                                beta_dup  == NULL ? beta  : beta_dup,  c );
      else
         PLA_Symm_left_lower_pp(nb_alg,
                                alpha_dup == NULL ? alpha : alpha_dup, a, b,
                                beta_dup  == NULL ? beta  : beta_dup,  c );
   }
   else /* if( PLA_SIDE_RIGHT == side) */
   {
      if(PLA_UPPER_TRIANGULAR == uplo)
         PLA_Symm_right_upper_pp(nb_alg,
                                alpha_dup == NULL ? alpha : alpha_dup, a, b,
                                beta_dup  == NULL ? beta  : beta_dup,  c );
      else
         PLA_Symm_right_lower_pp(nb_alg,
                                alpha_dup == NULL ? alpha : alpha_dup, a, b,
                                beta_dup  == NULL ? beta  : beta_dup,  c );
   }
   PLA_Obj_free( &alpha_dup );
   PLA_Obj_free( &beta_dup  );
   return PLA_SUCCESS;
}
/*****************************************************************************/

