
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/





#include "PLA.h"



int pla_shift(PLA_Obj *mv, int shift)
{
  int nb;
  PLA_Template template;

  PLA_Obj_template(*mv, &template);
  PLA_Temp_nb(template, &nb);

  /* if the amount of shift is a multiple of a blocksize */

  if ((shift%nb) == 0) 
    pla_shift_mult_nb(mv, shift);
  else 
    pla_shift_via_mscalar(mv, shift);
  
  return(0);
}



int pla_shift_mult_nb(PLA_Obj *mv, int shift)
{
/* shifts the multivector with respect to the template.
the amount of the "shift" has to be a multiple of the block
size 

I think this is sufficient for this version of the library
philip alpatov
*/ 
  int 
    error =0,
    me,
    nb,np,
    row, col, nprows, npcols,
    proc_send_to, proc_recv_from, proc_shift,
    idummy,
    count_in, count_out,
    glob_mv_length,
    glob_mv_width,
    glob_mv_align, current_align,
    loc_mv_length,
    loc_new_mv_length, loc_new_mv_width,
    new_mv_lld, new_mv_stride;

  MPI_Datatype
    datatype;

  MPI_Comm
    comm;

  PLA_Template
    template;

  PLA_Obj
    new_mv = NULL;

  void
    *new_mv_buf = NULL,
    *sendbuffer = NULL;





  PLA_Obj_template(*mv, &template);
  PLA_Temp_nb(template, &nb);
  PLA_Temp_total_num_proc(template, &np);
  PLA_Temp_me1D(template, &me);
  PLA_Temp_comm1D(template, &comm);

  if ((shift%nb) != 0)
    {
      printf("shift has value that is not multiple of the blocksize\n" );
      return(++error);
    }


  error += PLA_Mvector_global_all(*mv, &glob_mv_length, &glob_mv_width,
				 &glob_mv_align, &template);
  error += PLA_Obj_datatype( *mv, &datatype);
  error += PLA_Obj_local_length( *mv, &loc_mv_length );
  error += PLA_Obj_local_buffer( *mv, &sendbuffer );


  error += PLA_Mvector_create(datatype,
			     glob_mv_length,
			     glob_mv_width,
			     template,
			     glob_mv_align + shift,
			     &new_mv);

  error += PLA_Mvector_local_all( new_mv, &datatype, &loc_new_mv_length,
				 &idummy, &new_mv_buf, &new_mv_stride,
				 &new_mv_lld);



  
  error += PLA_Temp_comm_row_size(template, &npcols);
  error += PLA_Temp_comm_col_size(template, &nprows);
    
  
  
  proc_shift = ((shift/nb)%np + 2*np)%np;
  proc_send_to = (me+proc_shift)%np;
  proc_recv_from = (me + np - proc_shift)%np;
  /*
    printf("me = %d, proc_shift = %d, proc_send_to = %d, proc_recv_from = %d\n",
	 me, proc_shift, proc_send_to, proc_recv_from);
	 */

  count_out = loc_mv_length*glob_mv_width;
  count_in = loc_new_mv_length*glob_mv_width;

  error += MPI_Send(sendbuffer, count_out, datatype, proc_send_to, 
		    1, comm);

  error += MPI_Recv(new_mv_buf, count_in,
		    datatype, proc_recv_from, 1, comm,
		    (MPI_Status *)&idummy);
				     
  error += PLA_Obj_swap(mv, &new_mv);
  error += PLA_Obj_free(&new_mv);


  if (error != 0)
    {
      printf("PLA_shift completed with error = %d\n", error);
    }


  return(error);
}


int pla_shift_via_mscalar( PLA_Obj *mv, int shift )
{
  /* shifting a multivector via two consecutive multiscalar coppies */

  int
    glob_mv_length, glob_mv_width,
    glob_mv_align;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    temp_ms = NULL,
    new_mv = NULL;



  PLA_Mvector_global_all(*mv, &glob_mv_length, &glob_mv_width,
			 &glob_mv_align, &template);
  PLA_Obj_datatype( *mv, &datatype);

  PLA_Mvector_create(datatype,
		     glob_mv_length,
		     glob_mv_width,
		     template,
		     glob_mv_align + shift,
		     &new_mv);
  
  PLA_Mscalar_create_conf_to( *mv, 0,0, &temp_ms );
  PLA_Copy_mvector_to_mscalar( *mv, temp_ms );

  PLA_Copy_mscalar_to_mvector( temp_ms, new_mv ); 
  PLA_Obj_swap( mv, &new_mv );
  PLA_Obj_free( &new_mv );
  PLA_Obj_free( &temp_ms );
  return(0);
}



