


/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/





#ifndef PLA_MSCALAR_H
#define PLA_MSCALAR_H


#ifndef NULL
#define NULL 0
#endif


/* PLA_Mscalar_create
   Creates a multiscalar.  Subsequently will be freed by PLA_Obj_free.
  
  This routine sill allocate a buffer large enough to hold the multiscalar
  of lenght global_length and datatype.
*/
int PLA_Mscalar_create (MPI_Datatype      datatype, 
                       int               owner_row,
                       int               owner_col,
		       int               global_length,
		       int               global_width,
		       PLA_Template       template,
		       PLA_Obj          *new_mscalar);

/* PLA_Mscalar_submscalar
  This routines allocates a new multiscalar object, but does not allocate any memory
 to hold the data.  Instead it just contains a pointer to the buffer allocated
 in PLA_Multiscalar_create.  Hence changing this buffer will change the origional
 buffer.

 Also once a submultiscalar is created, freeing the origional multiscalar 
 will NOT result in freeing the submultiscalar or the buffer origionally 
 allocated.  Instead all memory will be preserved untill all submultiscalars 
 and the origional multiscalar have been freed with PLA_Obj_free, 
 at which time the master buffer will be freed.
*/
int PLA_Mscalar_submscalar (PLA_Obj old_mscalar, 
			   int      global_length, 
			   int      global_width,
			   int      mscalar_align_row,
			   int      mscalar_align_col,
			   PLA_Obj *new_mscalar);

/* PLA_Mscalar_local_all */
int PLA_Mscalar_local_all (PLA_Obj          Mscalar,
			  MPI_Datatype *    datatype,
			  int *             local_length,
			  int *             local_width,
			  void **           buffer,
			  int  *            local_lld);

/* PLA_Mscalar_global_all */
int PLA_Mscalar_global_all (PLA_Obj          Mscalar,
			   PLA_Template *     template);


/*  PLA_Mscalar_scalar_double
 NOTE: This routine should not be called by the user and is included here
 for the sake of completeness.  
 
 This routine is called only ONCE by PLA_Temp_create to set up some
 multiscalar constants that can then be used in calling the blas routines
 
 Two double precision multiscalars are created.  Each contain only one 
 element, a zero or a one.
*/
int PLA_Mscalar_scalar_double (PLA_Template    template,
			      PLA_Obj      *double_zero,
			      PLA_Obj      *double_one);
#endif



