/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/

#include "PLA.h"

int PLA_Chol ( int uplo, PLA_Obj a )
{
   int error = 0;

   PLA_Chol_enter ( uplo, a );

   switch ( uplo )
   {
      case PLA_UPPER_TRIANGULAR :  PLA_Chol_u ( a );
                                   break;
      case PLA_LOWER_TRIANGULAR :  PLA_Chol_l ( a );
                                   break;
      default                   :  break;
   }

   PLA_Chol_exit ( uplo, a );
   return ( PLA_SUCCESS );
}

int PLA_Chol_l ( PLA_Obj a )
{
   int size, nb, length, top_size, left_size, top_owner, left_owner, in_size;
   MPI_Datatype datatype;
   PLA_Template template;
   PLA_Obj a_relevent = NULL, a_left = NULL, a_right = NULL,
           a_10 = NULL, a_11 = NULL, a_20 = NULL, a_21 = NULL,
           a_10_dup = NULL, a_11_dup = NULL, a_21_dup = NULL,
           a_right_dup = NULL, work_row = NULL, work_col = NULL,
           top_a_10 = NULL, bottom_a_10 = NULL,
           top_a_10_dup = NULL, bottom_a_10_dup = NULL,
           top_a_11_dup = NULL, bottom_a_11_dup = NULL, right_a_11_dup = NULL,
           one = NULL, zero = NULL, min_one = NULL;

   PLA_Obj_template ( a, & template );
   PLA_Obj_datatype ( a, & datatype );

   PLA_Environ_nb_alg ( PLA_OP_MAT_PAN, template, & nb );

   PLA_Pmvector_create_conf_to ( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb, & work_row );
   PLA_Pmvector_create_conf_to ( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb, & work_col );

   PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &one );
   PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &zero );
   PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &min_one );

   PLA_Obj_set_to_one ( one );
   PLA_Obj_set_to_zero ( zero );
   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_view_all ( a, & a_relevent );
   PLA_Obj_global_length ( a_relevent, & length );

   while ( ( size = min ( length, nb ) ) > 0 )
   {
      PLA_Obj_vert_split_2 ( a_relevent, -length, & a_left, & a_right );
      PLA_Obj_view_shift ( a_right, 0, 0, size - length, 0 );
      PLA_Obj_horz_split_2 ( a_left,  size, & a_10, & a_20 );
      PLA_Obj_horz_split_2 ( a_right, size, & a_11, & a_21 );

      PLA_Obj_view_all ( work_row, & a_10_dup );
      PLA_Obj_view_shift ( a_10_dup, 0, 0, -length, size - nb );

      PLA_Obj_view_all ( work_col, & a_right_dup );
      PLA_Obj_view_shift ( a_right_dup, 0, 0, size - nb, 0 );
      PLA_Obj_horz_split_2 ( a_right_dup, size, & a_11_dup, & a_21_dup );

      PLA_Obj_set_to_zero ( a_right_dup );

      PLA_Obj_set_orientation ( a_10, PLA_PROJ_ONTO_ROW );
      PLA_Copy ( a_10, a_10_dup );


      PLA_Obj_view_all ( a_10, & bottom_a_10 );
      PLA_Obj_view_all ( a_10_dup, & bottom_a_10_dup );
      PLA_Obj_view_all ( a_11_dup, & right_a_11_dup );
      PLA_Obj_split_size ( right_a_11_dup, PLA_SIDE_TOP, & top_size, & top_owner );
      PLA_Obj_split_size ( right_a_11_dup, PLA_SIDE_LEFT, & left_size, & left_owner );

      while ( ( in_size = min ( top_size, left_size ) ) > 0 )
      {
         PLA_Obj_split_4 ( right_a_11_dup, in_size, in_size,
                           & top_a_11_dup,    PLA_DUMMY,
                           & bottom_a_11_dup, & right_a_11_dup );
         PLA_Obj_horz_split_2 ( bottom_a_10, in_size, & top_a_10,
                                                      & bottom_a_10 );
         PLA_Obj_horz_split_2 ( bottom_a_10_dup, in_size, & top_a_10_dup,
                                                          & bottom_a_10_dup );

         PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS, min_one, top_a_10,
                                                       zero,    top_a_11_dup );
         PLA_Local_gemm ( PLA_NO_TRANS, PLA_TRANS, min_one, bottom_a_10, top_a_10_dup,
                                                   zero,    bottom_a_11_dup );

         PLA_Obj_split_size ( right_a_11_dup, PLA_SIDE_TOP, & top_size, & top_owner );
         PLA_Obj_split_size ( right_a_11_dup, PLA_SIDE_LEFT, & left_size, & left_owner );
      }

      PLA_Local_gemm ( PLA_NO_TRANS, PLA_TRANS, min_one, a_20, a_10_dup,
                                                zero,    a_21_dup );

      PLA_Obj_set_orientation ( a_right, PLA_PROJ_ONTO_COL );

      PLA_Reduce_x ( PLA_SHAPE_LOW_TRAP, a_right_dup, one, a_right );

      PLA_Chol_l_nbtemp ( a_11 );

      PLA_Trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                 one, a_11, a_21 );

      PLA_Obj_view_shift ( a_relevent, size, 0, 0, 0 );
      PLA_Obj_view_shift ( work_col, size, 0, 0, 0 );
      PLA_Obj_global_length ( a_relevent, & length );
   }

   PLA_Obj_free ( & a_relevent );
   PLA_Obj_free ( & a_left );
   PLA_Obj_free ( & a_right );
   PLA_Obj_free ( & a_10 );
   PLA_Obj_free ( & a_20 );
   PLA_Obj_free ( & a_11 );
   PLA_Obj_free ( & a_21 );
   PLA_Obj_free ( & a_10_dup );
   PLA_Obj_free ( & a_11_dup );
   PLA_Obj_free ( & a_21_dup );
   PLA_Obj_free ( & a_right_dup );
   PLA_Obj_free ( & top_a_10 );
   PLA_Obj_free ( & bottom_a_10 );
   PLA_Obj_free ( & top_a_10_dup );
   PLA_Obj_free ( & bottom_a_10_dup );
   PLA_Obj_free ( & top_a_11_dup );
   PLA_Obj_free ( & bottom_a_11_dup );
   PLA_Obj_free ( & right_a_11_dup );
   PLA_Obj_free ( & work_row );
   PLA_Obj_free ( & work_col );
   PLA_Obj_free ( & one );
   PLA_Obj_free ( & zero );
   PLA_Obj_free ( & min_one );

   return ( PLA_SUCCESS );
}


int PLA_Chol_l_nbtemp ( PLA_Obj a )
{
   int nb, size, top_size, top_owner, left_size, left_owner, my_col;
   MPI_Datatype datatype;
   PLA_Template template;
   PLA_Obj a_00 = NULL, a_10 = NULL, a_11 = NULL, one = NULL, min_one = NULL,
           work_mscal = NULL, a_00_dup = NULL,
           work_row = NULL, work_col = NULL,
           dup_col_top = NULL, dup_col_bottom = NULL,
           dup_row_left = NULL, dup_row_right = NULL,
           topleft = NULL, bottomleft = NULL, bottomright = NULL;

   PLA_Obj_datatype ( a, & datatype );
   PLA_Obj_template ( a, & template );
   PLA_Temp_nb ( template, & nb );
   PLA_Temp_comm_row_rank ( template, & my_col );

   PLA_Pmvector_create_conf_to ( a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, nb, & work_row );
   PLA_Pmvector_create_conf_to ( a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, nb, & work_col );

   PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, nb, nb, template, & work_mscal );
   PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, & one );
   PLA_Mscalar_create ( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, & min_one );
   PLA_Obj_set_to_one ( one );
   PLA_Obj_set_to_minus_one ( min_one );

   PLA_Obj_view_all ( a, & a_11 );
   PLA_Obj_split_size ( a_11, PLA_SIDE_TOP, & top_size, & top_owner );
   PLA_Obj_split_size ( a_11, PLA_SIDE_LEFT, & left_size, & left_owner );

   while ( ( size = min3 ( nb, top_size, left_size ) ) > 0 )
   {
      PLA_Obj_split_4 ( a_11, size, size, & a_00, PLA_DUMMY,
                                          & a_10, & a_11 );

      PLA_Obj_view_all ( work_mscal, & a_00_dup );
      PLA_Obj_view_shift ( a_00_dup, 0, 0, size - nb, size - nb );

      PLA_Local_chol ( a_00 );

      PLA_Obj_objtype_cast ( a_00, PLA_MSCALAR );
      PLA_Copy ( a_00, a_00_dup );

      PLA_Local_trsm ( PLA_SIDE_RIGHT, PLA_LOW_TRIAN, PLA_TRANS, PLA_NONUNIT_DIAG,
                       one, a_00_dup, a_10 );
      PLA_Obj_view_shift ( work_col, size, 0, 0, 0 );
      PLA_Obj_view_shift ( work_row, 0, size, 0, 0 );

      PLA_Obj_view_all ( work_col, & dup_col_bottom );
      PLA_Obj_view_all ( work_row, & dup_row_right );

      PLA_Obj_view_shift ( dup_col_bottom, 0, 0, size - nb, 0 );
      PLA_Obj_view_shift ( dup_row_right, 0, 0, 0, size - nb );

      PLA_Copy ( a_10, dup_col_bottom );
      PLA_Copy ( a_10, dup_row_right );

      PLA_Obj_view_all ( a_11, & bottomright );

      PLA_Obj_split_size ( bottomright, PLA_SIDE_TOP, & top_size, & top_owner );
      PLA_Obj_split_size ( bottomright, PLA_SIDE_LEFT, & left_size, & left_owner );

      while ( ( size = min3 ( nb, top_size, left_size ) ) > 0 )
      {
         PLA_Obj_split_4 ( bottomright, size, size, & topleft, PLA_DUMMY,
                                                    & bottomleft, & bottomright );

         PLA_Obj_vert_split_2 ( dup_row_right, size, & dup_row_left, & dup_row_right );
         PLA_Obj_horz_split_2 ( dup_col_bottom, size, & dup_col_top, & dup_col_bottom );

         PLA_Local_syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS,
                          min_one, dup_col_top, one, topleft );

         PLA_Local_gemm ( PLA_NO_TRANS, PLA_NO_TRANS,
                          min_one, dup_col_bottom, dup_row_left, one, bottomleft );

         PLA_Obj_split_size ( bottomright, PLA_SIDE_TOP, & top_size, & top_owner );
         PLA_Obj_split_size ( bottomright, PLA_SIDE_LEFT, & left_size, & left_owner );
      }


      PLA_Obj_split_size ( a_11, PLA_SIDE_TOP, & top_size, & top_owner );
      PLA_Obj_split_size ( a_11, PLA_SIDE_LEFT, & left_size, & left_owner );

   }

   PLA_Obj_free ( & a_00 );
   PLA_Obj_free ( & a_10 );
   PLA_Obj_free ( & a_11 );
   PLA_Obj_free ( & work_mscal );
   PLA_Obj_free ( & a_00_dup );
   PLA_Obj_free ( & work_row );
   PLA_Obj_free ( & work_col );
   PLA_Obj_free ( & dup_col_top );
   PLA_Obj_free ( & dup_col_bottom );
   PLA_Obj_free ( & dup_row_left );
   PLA_Obj_free ( & dup_row_right );
   PLA_Obj_free ( & topleft );
   PLA_Obj_free ( & bottomleft );
   PLA_Obj_free ( & bottomright );
   PLA_Obj_free ( & one );
   PLA_Obj_free ( & min_one );

   return ( PLA_SUCCESS );
}

int PLA_Chol_u ( PLA_Obj a )
{
   return ( PLA_SUCCESS );
}

int PLA_Local_chol ( PLA_Obj Aobj )
{
   int loc_len, loc_wid, ldim;
   void * buffer;
   MPI_Datatype datatype;

   PLA_Obj_local_length ( Aobj, & loc_len );
   PLA_Obj_local_width ( Aobj, & loc_wid );

   if ( loc_len > 0 && loc_wid > 0 )
   {
      if ( loc_len != loc_wid )
      {
         printf ( "Non-square dimensions in PLA_Local_chol " );
         printf ( ": length = %d, width = %d\n", loc_len, loc_wid );
         return ( -1 );
      }
      
      PLA_Obj_local_ldim ( Aobj, & ldim );
      PLA_Obj_datatype ( Aobj, & datatype );
      PLA_Obj_local_buffer ( Aobj, & buffer );

      if ( datatype == MPI_DOUBLE )
         PLA_dchol ( loc_len, ( double * ) buffer, ldim );
      else if ( datatype == MPI_FLOAT )
         PLA_schol ( loc_len, ( float * ) buffer, ldim );
      else
      {
         printf ( "Invalid datatype in PLA_Local_chol\n" );
         return ( -1 );
      }
   }
   return ( 0 );
}

int PLA_dchol ( int n, double * buffer, int lld )
{
   int k,
       length,
       int_one = 1;
   double invert,
          min_one = -1.0e+0,
          one = 1.0e+0;

#define a(i,j) buffer [ ( j ) * lld + ( i ) ]

   for ( k = 0 ; k < n ; k++ )
   {
      if ( k > 0 )
      {
         length = n - k;
         PLA_dgemv ( "N", & length, & k,
                     & min_one, & a ( k, 0 ), & lld, & a ( k, 0 ), & lld,
                     & one, & a ( k, k ), & int_one );
      }

      a ( k, k ) = sqrt ( a ( k, k ) );

      length = n - k - 1;

      if ( length > 0 )
      {
         invert = one / a ( k, k );
         PLA_dscal ( & length, & invert, & a ( k + 1, k ), & int_one );
      }
   }

#undef a

   return ( 0 );
}

int PLA_schol ( int n, float * buffer, int lld )
{
   int k,
       length,
       int_one = 1;
   float invert,
         min_one = -1.0e+0,
         one = 1.0e+0;

#define a(i,j) buffer [ ( j ) * lld + ( i ) ]

   for ( k = 0 ; k < n ; k++ )
   {

      if ( k > 0 )
      {
         length = n - k;
         PLA_sgemv ( "N", & length, & k,
                     & min_one, & a ( k, 0 ), & lld, & a ( k, 0 ), & lld,
                     & one, & a ( k, k ), & int_one );
      }

      a ( k, k ) = sqrt ( a ( k, k ) );

      length = n - k - 1;

      if ( length > 0 )
      {
         invert = one / a ( k, k );
         PLA_sscal ( & length, & invert, & a ( k + 1, k ), & int_one );
      }
   }

#undef a

   return ( 0 );
}
