

/* PLA_Local_gemm.c */

#include "PLA.h"
#include <stdio.h>

int PLA_Local_gemm( int trans_A,
						 int trans_B,
						 PLA_Obj alpha,
						 PLA_Obj A,
						 PLA_Obj B,
						 PLA_Obj beta,
						 PLA_Obj C )
{
	int error=0;
	int local_length, check_length;
	int local_width , check_width;
	int local_inner;
	int lda, ldb, ldc;
	MPI_Datatype Dtype;
	void *Abuf, *Bbuf, *Cbuf;
	void *alphabuf, *betabuf;
	char Trans_A[1], Trans_B[1];


	PLA_Obj_local_buffer( A, &Abuf);
	PLA_Obj_local_buffer( B, &Bbuf);
	PLA_Obj_local_buffer( C, &Cbuf);

	PLA_Obj_local_buffer( alpha, &alphabuf);
	PLA_Obj_local_buffer( beta,  &betabuf);

	if(PLA_NO_TRANS == trans_A) 
	{
		Trans_A[0] = 'N';
		PLA_Obj_local_width(  A, &local_inner);
		PLA_Obj_local_length( A, &local_length);
	}
	else if( PLA_TRANS == trans_A )
	{
		Trans_A[0] = 'T';
		PLA_Obj_local_length(  A, &local_inner);
		PLA_Obj_local_width( A, &local_length);
	}
	else if(PLA_CONJ == trans_A) 
	{
		Trans_A[0] = 'N';
		PLA_Obj_local_width(  A, &local_inner);
		PLA_Obj_local_length( A, &local_length);
                PLA_Conjugate(A);
	}
	else /* if( PLA_CONJ_TRANS == trans_A ) */
	{
		Trans_A[0] = 'C';
		PLA_Obj_local_length(  A, &local_inner);
		PLA_Obj_local_width( A, &local_length);
	}

	if (PLA_NO_TRANS == trans_B) 
	{
		 Trans_B[0] = 'N';
		PLA_Obj_local_width( B, &local_width);
	}
	else if( PLA_TRANS == trans_B)
	{
		Trans_B[0] = 'T';
		PLA_Obj_local_length( B, &local_width);
	}
	else if (PLA_CONJ == trans_B) 
	{
		Trans_B[0] = 'N';
		PLA_Obj_local_width( B, &local_width);
                PLA_Conjugate(B);
	}
	else /* if( PLA_CONJ_TRANS == trans_A ) */
	{
		Trans_B[0] = 'C';
		PLA_Obj_local_length( B, &local_width);
	}

	PLA_Obj_local_ldim ( A, &lda);
	PLA_Obj_local_ldim ( B, &ldb); 
	PLA_Obj_local_ldim ( C, &ldc);

	PLA_Obj_datatype(C, &Dtype);

	if (0 != local_length  &&
		 0 != local_width   &&
		 0 != local_inner)
	{
		if(Dtype == MPI_DOUBLE)
				  PLA_dgemm(Trans_A,Trans_B,&local_length,&local_width,&local_inner,
									alphabuf, Abuf, &lda, Bbuf, &ldb, betabuf, Cbuf, &ldc);
		else if(Dtype == MPI_FLOAT)
				  PLA_sgemm(Trans_A,Trans_B,&local_length,&local_width,&local_inner,
									alphabuf, Abuf, &lda, Bbuf, &ldb, betabuf, Cbuf, &ldc);
		else if(Dtype == MPI_COMPLEX)
				  PLA_cgemm(Trans_A,Trans_B,&local_length,&local_width,&local_inner,
									alphabuf, Abuf, &lda, Bbuf, &ldb, betabuf, Cbuf, &ldc);
		else if(Dtype == MPI_DOUBLE_COMPLEX)
				  PLA_zgemm(Trans_A,Trans_B,&local_length,&local_width,&local_inner,
									alphabuf, Abuf, &lda, Bbuf, &ldb, betabuf, Cbuf, &ldc);

		 else error ++;
	}
return(error);
}
