
/***************************************************************************
  20 June 1996 1:00pm
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by: 
  Robert van de Geijn, Department of Computer Sciences
  University of Texas at Austin  78712.    rvdg@cs.utexas.edu

  Philip A. Alpatov Department of Physics, UT Austin.
  philip@physics.utexas.edu

***************************************************************************/


#include "PLA.h"
#include "PLA_BLAS.h"
#include "MISC_DEFS.h"
/*
 * PLA_Local_gemv
 */


int PLA_Local_gemv( int      trans,
		    PLA_Obj alpha,
		    PLA_Obj a,
		    PLA_Obj x,
		    PLA_Obj beta,
		    PLA_Obj y )
/* 
   compute local part of
          
                 y = alpha A x + y

   Currently supported cases:
   x,y      VECTOR   ROW/COL VECTOR
   A        matrix
   alpha,
       beta MScalar (1x1)
*/                 
{
  MPI_Datatype datatype_alpha, 
               datatype_x,
               datatype_a,
               datatype_y; 
  int          type_x, 
               type_y,
               local_length_a,
               local_width_a,
               local_length_x,
               local_length_y,
	       local_stride_x,
	       local_stride_y,
               a_lld,
               error;
  void        *buffer_x,
              *buffer_y,
              *buffer_a,
              *buffer_alpha,
              *buffer_beta,
              *buffer_lld;
  char         trans_char;


  trans_char = (  PLA_NO_TRANSPOSE == trans ? 'N' :
			(PLA_TRANSPOSE    == trans ? 'T' : 'C'));

  PLA_Matrix_local_all( a, &datatype_a, &local_length_a, &local_width_a,
				&buffer_a, &a_lld);
  PLA_Vector_local_all( x, &datatype_x, &local_length_x,
						&buffer_x, &local_stride_x );
  PLA_Vector_local_all( y, &datatype_y, &local_length_y,
				 &buffer_y, &local_stride_y );
  PLA_Obj_local_buffer(alpha, &buffer_alpha);
  PLA_Obj_local_buffer(beta, &buffer_beta);

  if( local_width_a > 0 && local_length_a > 0)
  {
	  if ( MPI_FLOAT == datatype_a )
		  PLA_sgemv ( &trans_char, &local_length_a, &local_width_a,
							buffer_alpha, buffer_a, &a_lld,
							buffer_x, &local_stride_x, buffer_beta,
							buffer_y, &local_stride_y );
	  else if (MPI_DOUBLE == datatype_a)
		  PLA_dgemv ( &trans_char, &local_length_a, &local_width_a,
						  buffer_alpha, buffer_a, &a_lld,
						  buffer_x, &local_stride_x, buffer_beta,
						  buffer_y, &local_stride_y );
	  else if (MPI_COMPLEX == datatype_a)
		  PLA_cgemv ( &trans_char, &local_length_a, &local_width_a,
						  buffer_alpha, buffer_a, &a_lld,
						  buffer_x, &local_stride_x, buffer_beta,
						  buffer_y, &local_stride_y );
	  else if (MPI_DOUBLE_COMPLEX == datatype_a)
		  PLA_zgemv ( &trans_char, &local_length_a, &local_width_a,
						  buffer_alpha, buffer_a, &a_lld,
						  buffer_x, &local_stride_x, buffer_beta,
						  buffer_y, &local_stride_y );
  }
  else
	  PLA_Obj_set_to_zero(y);
}

