/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/



#include "PLA.h"

int  PLA_Local_trmm(int side, int uplo, int trans, int unit,
                    PLA_Obj alpha, PLA_Obj  A, PLA_Obj B)
{
   void *alpha_buf, *a_buf, *b_buf;
   int  length_A, width_A, width_B;
   int  lda, ldb, m, n;
   char  side_char[1], uplo_char[1], trans_char[1], unit_char[1];
   MPI_Datatype datatype;

   PLA_Obj_datatype(A, &datatype);
   PLA_Obj_local_buffer(alpha, &alpha_buf);
   PLA_Obj_local_buffer(A, &a_buf);
   PLA_Obj_local_buffer(B, &b_buf);


   PLA_Obj_local_length(B, &m);
   PLA_Obj_local_width (B, &n);
   PLA_Obj_local_ldim(A, &lda);
   PLA_Obj_local_ldim(B, &ldb);
   if(PLA_SIDE_LEFT == side)
   	side_char[0] = 'L';
   else side_char[0] = 'R';
   if(PLA_UPPER_TRIANGULAR == uplo)
   	uplo_char[0] = 'U';
   else uplo_char[0] = 'L';
   if(PLA_NO_TRANSPOSE == trans)
   	trans_char[0] = 'N';
   else trans_char[0] = 'T';
   if(PLA_UNIT_DIAG == unit)
   	unit_char[0] = 'U';
   else unit_char[0] = 'N';
   if(m*n > 0)
   {
  if(MPI_DOUBLE == datatype)
  PLA_dtrmm (side_char, uplo_char, trans_char, unit_char, &m, &n, 
            (double*)alpha_buf,
            (double*)a_buf, &lda,
            (double*)b_buf, &ldb);
  else if (MPI_FLOAT == datatype)
  PLA_strmm (side_char, uplo_char, trans_char, unit_char, &m, &n, 
            (float*)alpha_buf,
            (float*)a_buf, &lda,
            (float*)b_buf, &ldb);
  else if (MPI_COMPLEX == datatype)
  PLA_ctrmm (side_char, uplo_char, trans_char, unit_char, &m, &n, 
            (PLA_COMPLEX*)alpha_buf,
            (PLA_COMPLEX*)a_buf, &lda,
            (PLA_COMPLEX*)b_buf, &ldb);
  else if (MPI_DOUBLE_COMPLEX == datatype)
  PLA_ztrmm (side_char, uplo_char, trans_char, unit_char, &m, &n, 
            (PLA_DOUBLE_COMPLEX*)alpha_buf,
            (PLA_DOUBLE_COMPLEX*)a_buf, &lda,
            (PLA_DOUBLE_COMPLEX*)b_buf, &ldb);
  }
}

