
#include "PLA_Local_BLAS.h"




int PLA_Local_trsv ( int uplo, int trans, int diag, PLA_Obj Aobj, PLA_Obj Xobj )
{
   MPI_Datatype        Adatatype,   Xdatatype;
   int                 Aloc_len,   Xloc_len,
                       Aloc_wid,   Xloc_wid,
                       Astride,    Xstride,
                       Aldim,      Xldim;
   void              * Abuffer,   * Xbuffer;
   char c_uplo, c_trans, c_diag;

   PLA_Obj_datatype (Aobj, & Adatatype);
   PLA_Obj_datatype (Xobj, & Xdatatype);

   PLA_Obj_local_info ( Aobj, & Aloc_len, & Aloc_wid,
                               & Abuffer, & Astride, & Aldim );
   PLA_Obj_local_info ( Xobj, & Xloc_len, & Xloc_wid,
                              & Xbuffer, & Xstride, & Xldim );


   if ( Adatatype != Xdatatype )
   {
      printf ( "Datatypes not consistant in PLA_Local_trsv!!!\n" );
      return ( -1 );
   }

   if ( Aloc_len > 0 && Aloc_wid > 0 )
   {
      if ( Aloc_len != Aloc_wid )
      {
         printf ( "Non-square dimensions in PLA_Local_trsv " );
         printf ( ": length = %d, width = %d\n", Aloc_len, Aloc_wid );
         return ( -1 );
      }

      if ( Aloc_len != Xloc_len )
      {
         printf ( "Uncompatable dimensions in PLA_Local_trsv " );
         printf ( ": A ( %d x %d ), X ( %d )\n", Aloc_len, Aloc_wid, Xloc_len );
         return ( -1 );
      }

      c_uplo = 'L';
      c_trans = 'N';
      c_diag = 'N';
      if ( uplo == PLA_UPPER_TRIANGULAR ) c_uplo = 'U';
      if ( trans == PLA_TRANSPOSE ) c_trans = 'T';

      if ( Adatatype == MPI_DOUBLE )
         PLA_dtrsv ( & c_uplo, & c_trans, & c_diag,
                      & Aloc_len, ( double * ) Abuffer, & Aldim,
                                  ( double * ) Xbuffer, & Xstride );
      else if ( Adatatype == MPI_FLOAT )
         PLA_strsv ( & c_uplo, & c_trans, & c_diag,
                      & Aloc_len, ( float * ) Abuffer, & Aldim,
                                  ( float * ) Xbuffer, & Xstride );
      else
      {
         printf ( "Invalid datatype in PLA_Local_trsv\n" );
         return ( -1 );
      }
   }
   return ( 0 );
}
