/* ---------------------------------------------------------------------
Author:     H. Carter Edwards
		http://www.ticam.utexas.edu/~carter
		carter@ticam.utexas.edu

Copyright:  Copyright (C) 1996   H. Carter Edwards

Purpose:    Asynchronous message management for MPI

License:    See 'License.Gnu'

Mail:
	University of Texas at Austin
	TICAM, TAY 2.400
	2313 Speedway
	Austin, TX  78712

Disclaimer:

    MMPI comes with ABSOLUTELY NO WARRANTY; for details read
    the file License.Gnu that came with this library.
    This is free software, and you are welcome to redistribute it
    under certain conditions, i.e. read the file 'License.Gnu'.
--------------------------------------------------------------------- */

#include "mmpi.h"

int main_test( MPI_Comm comm )
{
  int R , i ;
  double T1 , T2 ;

  if ( MPI_SUCCESS != ( R = MMPI_Enable( comm ) ) ) {
    MPI_Abort( comm , R );
  }

  fprintf(MMPI_Log_file(),"main_test: MMPI_Barrier #1\n" );

  T1 = MMPI_Wait_idle_time();

  if ( MPI_SUCCESS != ( R = MMPI_Barrier( comm ) ) )
    MPI_Abort( comm , R );

  T2 = MMPI_Wait_idle_time();

  fprintf(MMPI_Log_file(),"main_test: Time waiting = %f\n", T2 - T1 );

  fprintf(MMPI_Log_file(),"main_test: MMPI_Barrier #2\n" );

  T1 = MMPI_Wait_idle_time();

  if ( MPI_SUCCESS != ( R = MMPI_Barrier( comm ) ) )
    MPI_Abort( comm , R );

  T2 = MMPI_Wait_idle_time();

  fprintf(MMPI_Log_file(),"main_test: Time waiting = %f\n", T2 - T1 );

  fprintf(MMPI_Log_file(),"main_test: MMPI_Barrier #3\n" );

  T1 = MMPI_Wait_idle_time();

  if ( MPI_SUCCESS != ( R = MMPI_Barrier( comm ) ) )
    MPI_Abort( MPI_COMM_WORLD , R );

  T2 = MMPI_Wait_idle_time();

  fprintf(MMPI_Log_file(),"main_test: Time waiting = %f\n", T2 - T1 );

  MMPI_Disable( MPI_COMM_WORLD );

  return 0 ;
}


main( int argc , char * argv[] )
{
  int N = 5 ;
  int R , i ;

  for ( i = 0 ; i < argc ; ++i ) printf(" %s", argv[i] );
  printf("\n");

  if ( MPI_SUCCESS != ( R = MPI_Init( &argc , &argv ) ) ) {
    MPI_Abort( MPI_COMM_WORLD , R );
  }

  for ( i = 0 ; i < N ; ++i ) {
    fprintf(MMPI_Log_file(),"\nmain pass #%d\n", i );
    main_test( MPI_COMM_WORLD );
  }

  MPI_Finalize();

  return 0 ;
}

