


/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/


/* The routines PLA_Temp_first_col_block and PLA_Temp_first_row_block
   have been changed to accomodate the case when the alignment
   of the matrix is not a multiple of a blocksize 
   
   Aug 22 1996 Philip Alpatov philip@physics.utexas.edu */

#include "PLA.h"


/*
   comm should be a communicator with the two dimentional topology
   already defined.   It will probably be created with MPI_Cart_create.
   */


#ifndef min
#define min(x,y) (x <= y ? x : y)
#endif


int PLA_Temp_create (int            nb, 
		     PLA_Start_num   zero_or_one,
		     PLA_Template   *template_ret)

{
  int      dims[2], periods[2], coords[2];
  int      i,j,k;
  int      remain_dims[2];
  int      nnodes;
  int      error;
  int      num_of_rows, my_cvect_index;
  MPI_Comm       comm;
  MPI_Group world_group, temp_group;
  PLA_Template   template;

  dims[0]=dims[1]=0;

  /* 
     Allocate Memory for the template.  
     */
  template = (PLA_Template) pla_malloc (sizeof(PLA_Template_def));
  if (template==NULL) return(1);

  /* 
     Get the global communicator.
     */
  PLA_Base_comm (&comm);
  if (MPI_COMM_NULL==comm) {
     printf (" PLA_Init not called before PLA_Temp_create.\n");
     printf (" No global Communicator defined.\n");
  }

  /* 
     Store the parameters passed.
     */
  template->nb       = nb;
  template->comm2D   = comm;
  if (zero_or_one != PLA_ZERO  &&
      zero_or_one != PLA_ONE) {
    printf ("Error in PLA_temp_create: PLA_Start_num must be zero or one");
    printf (" found %d", zero_or_one);
    return (1);
  }
  template->zero_or_one = zero_or_one;


  /* 
     Set the other 2 communicators, row, column, and 1D.
     remain_dims tells which dimensions to keep and which to throw
     away.  For instance remain_dims[0]=1, remain_dims[1]=0 keeps the 
     rows and throws away the columns createing a col communicator.
     */
  
  remain_dims[0]=1;
  remain_dims[1]=0;
  error = MPI_Cart_sub (template->comm2D, 
			remain_dims, 
			&(template->commcol)); 
  pla_MPI_error (error, "MPI_Cart_sub") ;
  if (error) return(error);

  remain_dims[0]=0;
  remain_dims[1]=1;
  error = MPI_Cart_sub(template->comm2D, 
		       remain_dims, &
		       (template->commrow));
  pla_MPI_error (error, "MPI_Cart_sub") ;
  if (error) return(error);

  /* 
     To compute the row dimention, look at the size of the column 
     communicator.  To compute the size of the column dimentin
     look at the size of the row communicator.
     */
  error = MPI_Comm_size (template->commcol, 
			 &(template->num_proc_row));
  if (error) {
    printf ("Error returned from MPI_Comm_size\n");
    return(1);
  }
  error = MPI_Comm_size (template->commrow, 
			 &(template->num_proc_col));
  if (error) {
    printf ("Error returned from MPI_Comm_size\n");
    return(1);
  }
  /* 
     Store the position of this node within the processing grid.
     The postion in the row communicator is the column index
     and vice versa.
     */
  error = MPI_Comm_rank(template->commcol, 
			&(template->myrow));
  if (error) {
    printf ("Error returned from MPI_Comm_rank\n");
    return(1);
  }
  error = MPI_Comm_rank(template->commrow, 
			&(template->mycol));
  if (error) {
    printf ("Error returned from MPI_Comm_rank\n");
    return(1);
  } 

  error = PLA_Base_comm_1d (&template->comm1D);
  if (error) {
    printf ("Error returned from PLA_Base_comm_1d\n");
    return(1);
  }

  error = MPI_Comm_rank(template->comm1D, 
			&(template->me1D));
  if (error) {
    printf ("Error returned from MPI_Comm_rank\n");
    return(1);
  }
  
  /* 
     Now set the number of times this template has been used.  
     This number will be incremented when a matrix is allocated
     using this template.  It will be decremented when that
     matrix is realeased.  The
     template will not be released untill this number has been
     decremented to zero.

     If this system fails there is a stronger system that is not too
     hard to implement.  Based on ideas by Greg Morrow.
     Allocate the integer below and also have a simple flag that says
     if the template itself has been deallocated.  Only deallocate 
     when both flags are down to zero.  This defeats extracting
     the template out of a matrix structure and then freeing the
     template.  Once the matrix structure has been freed it is impossible
     to extract the template from it.
     */
  template->instances = (int *) pla_malloc(sizeof(int));
  *(template->instances) = 1;

  *template_ret = template;

  return(0);
}



int  PLA_Temp_free(PLA_Template *template)

{
  /* I guess check for NULL just to be safe.
     */
  if (NULL == *template) return(0);
  
  (*(*template)->instances)--;
  if ( *(*template)->instances == 0) {
    MPI_Comm_free (&((*template)->commcol));
    MPI_Comm_free (&((*template)->commrow));
    pla_free ((*template)->instances);
    
    pla_free (*template);
  }
  *template = NULL;
  return (0);
}



int PLA_Temp_comm_col_size(PLA_Template template, 
			 int *num_proc_row)
{
  *num_proc_row = (template->num_proc_row);
  return(0);
}

int PLA_Temp_comm_row_size(PLA_Template template,
			 int *num_proc_col)
{
  *num_proc_col = (template->num_proc_col);
  return(0);
}

int PLA_Temp_total_num_proc(PLA_Template template,
			   int *num_proc)
{
  *num_proc = (template->num_proc_col)*(template->num_proc_row);
  return(0);
}

int PLA_Temp_nb(PLA_Template template, 
	       int *nb)
{
  *nb = (template->nb);
  return(0);
}

int PLA_Temp_comm_all_info(PLA_Template template,
                           MPI_Comm    *comm,
                           int         *rank,
                           int         *numnodes)
{
   int return_value = 0;

   return_value += PLA_Temp_comm1D(template, comm);
   return_value += PLA_Temp_me1D(template, rank);
   return_value += PLA_Temp_total_num_proc(template, numnodes);

   return(return_value);
}

int PLA_Temp_comm_all(PLA_Template template,
                      MPI_Comm    *comm)
{
   int return_value = 0;

   return_value += PLA_Temp_comm1D(template, comm);

   return(return_value);
}

int PLA_Temp_comm_all_rank(PLA_Template template,
                           int         *rank)
{
   int return_value = 0;

   return_value += PLA_Temp_me1D(template, rank);

   return(return_value);
}

int PLA_Temp_comm_all_size(PLA_Template template,
                           int         *numnodes)
{
   int return_value = 0;

   return_value += PLA_Temp_total_num_proc(template, numnodes);

   return(return_value);
}


int PLA_Temp_comm_row(PLA_Template template, 
		    MPI_Comm *commrow)
{
  *commrow = (template->commrow);
  return(0);
}

int PLA_Temp_comm_row_info(PLA_Template   template,
                      MPI_Comm     *comm_row,
                      int          *rank_within_row,
                      int          *numnodes_in_row)
{
   int return_value = 0;

   return_value += PLA_Temp_comm_row(template, comm_row);
   return_value += PLA_Temp_comm_row_rank(template, rank_within_row);
/* Note : Here I cheat and/or abuse the naming convention in PLA */
/*        The number of processors IN a row is the number of    */
/*        processor columns.  John Gunnels                      */
   return_value += PLA_Temp_comm_row_size(template, numnodes_in_row);
   return(return_value);
}


int PLA_Temp_comm_col(PLA_Template template, 
		    MPI_Comm *commcol)
{
  *commcol = (template->commcol);
  return(0);
}


int PLA_Temp_comm_col_info(PLA_Template  template,
                      MPI_Comm     *comm_col,
                      int          *rank_within_col,
                      int          *numnodes_in_col)
{
   int return_value = 0;

   return_value += PLA_Temp_comm_col(template, comm_col);
   return_value += PLA_Temp_comm_col_rank(template, rank_within_col);
/* Note : Here I cheat and/or abuse the naming convention in PLA */
/*        The number of processors IN a column is the number of */
/*        processor rows.  John Gunnles                         */
   return_value += PLA_Temp_comm_col_size(template, numnodes_in_col);
   return(return_value);
}

int PLA_Temp_comm1D(PLA_Template template, 
		   MPI_Comm *comm1D)
{
  *comm1D = (template->comm1D);
  return(0);
}

int PLA_Temp_comm_col_rank(PLA_Template template, 
		  int *myrow)
{
  *myrow = (template->myrow);
  return(0);
}

int PLA_Temp_comm_row_rank(PLA_Template template, 
		  int *mycol)
{
  *mycol = (template->mycol);
  return(0);
}

int PLA_Temp_me1D(PLA_Template template, 
		 int *me1D)
{
  *me1D = (template->me1D);
  return(0);
}

int PLA_Temp_mycoords (PLA_Template template , 
		      int *myrow, 
		      int *mycol)
{
  *myrow = (template->myrow);
  *mycol = (template->mycol);
  return(0);
}

int PLA_Temp_zero_or_one(PLA_Template template, 
			PLA_Start_num *zero_or_one)
{
  *zero_or_one = (template->zero_or_one);
  return(0);
}



int PLA_Temp_vector_local_length (PLA_Template template, 
				 int global_n, 
				 int global_i,
				 int *local_n) 

{
  int error=0;
  int me1D;
  
  error += PLA_Temp_me1D          (template, &me1D);
  error += PLA_Temp_vector_distant_length (template,
					  me1D,
					  global_n, 
					  global_i,
					  local_n);
  return (error);
}

int PLA_Temp_vector_distant_length (PLA_Template template,
				   int processor_num_1D,
				   int global_n, 
				   int global_i,
				   int *local_n) 
{
  int icur, num_procs, nb, error=0; 
  PLA_Start_num zero_or_one;

  error += PLA_Temp_total_num_proc(template, &num_procs);
  error += PLA_Temp_nb            (template, &nb);
  error += PLA_Temp_zero_or_one   (template, &zero_or_one);

  error += PLA_Temp_local_compute (global_n,
				  global_i, 
				  processor_num_1D,
				  nb,
				  num_procs,
				  zero_or_one,
				  local_n, 
				  &icur);
  return (error);
}


int PLA_Temp_vector_owner_rank (PLA_Template template, 
			       int global_i,
			       int *rank) 

{
  int myn, num_procs, error=0;
  int me1D, nb;
  PLA_Start_num zero_or_one;

  error += PLA_Temp_total_num_proc(template, &num_procs);
  error += PLA_Temp_me1D          (template, &me1D);
  error += PLA_Temp_nb            (template, &nb);
  error += PLA_Temp_zero_or_one   (template, &zero_or_one);

  /* 1 is the lenght of the vector we are finding who owns
     the beginning of if offset is global_i.
     */
  error += PLA_Temp_local_compute (1,
				  global_i, 
				  me1D,
				  nb,
				  num_procs,
				  zero_or_one,
				  &myn, rank);
  return (error);
}

int PLA_Temp_matrix_local_length (PLA_Template template, 
				 int global_m, 
				 int global_i,
				 int *local_m)
     
{
  int  error=0;
  int myrow;

  error += PLA_Temp_comm_col_rank       (template, &myrow),
  error + PLA_Temp_matrix_distant_length (template,
					 myrow,
					 global_m, 
					 global_i,
					 local_m);
  return (error);
} 


int PLA_Temp_matrix_local_width (PLA_Template template, 
				int global_n, 
				int global_j,
				int *local_n)
{
  int error=0;
  int mycol;

  error += PLA_Temp_comm_row_rank       (template, &mycol),
  error += PLA_Temp_matrix_distant_width (template, 
				         mycol,
					 global_n, 
					 global_j,
					 local_n);
  return (error);
} 

int PLA_Temp_matrix_local_length_width (PLA_Template template,
				       int global_m, 
				       int global_n, 
				       int global_i, 
				       int global_j, 
				       int *local_m, 
				       int *local_n)
{
  int error=0;
  error += PLA_Temp_matrix_local_length (template, 
					global_m, 
					global_i,
					local_m);

  error += PLA_Temp_matrix_local_width  (template, 
					global_n, 
					global_j, 
					local_n);
  return(error);
} 


int PLA_Temp_matrix_distant_length (PLA_Template template,
				   int dist_proc_row,
				   int global_m, 
				   int global_i,
				   int *distant_m)

{
  int myn, icur, error=0;
  int num_proc_rows, nb;
  PLA_Start_num zero_or_one;

  error += PLA_Temp_comm_col_size(template, &num_proc_rows),
  error += PLA_Temp_nb          (template, &nb),
  error += PLA_Temp_zero_or_one (template, &zero_or_one),  

  error += PLA_Temp_local_compute (global_m, 
				  global_i, 
				  dist_proc_row, 
				  nb, 
				  num_proc_rows, 
				  zero_or_one, 
				  distant_m, &icur);
  return (error);
} 


int PLA_Temp_matrix_distant_width (PLA_Template template, 
				  int distant_proc_col,
				  int global_n, 
				  int global_j,
				  int *distant_n)
{
  int myn, icur, error=0;
  int nb_col;
  int num_proc_rows, num_proc_cols, nb;
  PLA_Start_num zero_or_one;

  error += PLA_Temp_comm_col_size(template, &num_proc_rows);
  error += PLA_Temp_comm_row_size(template, &num_proc_cols);
  error += PLA_Temp_nb          (template, &nb);
  error += PLA_Temp_zero_or_one (template, &zero_or_one),  

  nb_col =  num_proc_rows * nb;
  error += PLA_Temp_local_compute (global_n,
				  global_j, 
				  distant_proc_col,
				  nb_col,
				  num_proc_cols,
				  zero_or_one,
				  distant_n, &icur);
  return (error);
} 

int PLA_Temp_matrix_distant_length_width (PLA_Template template,
					 int distant_proc_row,
					 int distant_proc_col,
					 int global_m, 
					 int global_n, 
					 int global_i, 
					 int global_j, 
					 int *distant_m, 
					 int *distant_n)
{
  int error=0;
  error += PLA_Temp_matrix_distant_length (template,
					  distant_proc_row,
					  global_m, 
					  global_i,
					  distant_m);

  error += PLA_Temp_matrix_distant_width (template,
					 distant_proc_col, 
					 global_n, 
					 global_j, 
					 distant_n);
  return(error);
} 



int PLA_Temp_matrix_owner_row (PLA_Template template,
			      int global_i,
			      int *rank) 

{
  int myn, icur, error=0;
  int myrow, nb, num_proc_rows;
  PLA_Start_num zero_or_one;

  error += PLA_Temp_comm_col_rank       (template, &myrow);
  error += PLA_Temp_nb          (template, &nb);
  error += PLA_Temp_comm_col_size(template, &num_proc_rows),
  error += PLA_Temp_zero_or_one (template, &zero_or_one),  
 
  error += PLA_Temp_local_compute (1,global_i, 
				  myrow,
				  nb,
				  num_proc_rows,
				  zero_or_one,
				  &myn, rank);
  return (error);
} 


int PLA_Temp_matrix_owner_col (PLA_Template template,
			      int global_j,
			      int *rank) 
{
  int myn, icur, error=0;
  int nb_col, np_col, nb;
  int mycol, num_proc_cols, num_proc_rows;
  PLA_Start_num zero_or_one;

  error += PLA_Temp_comm_col_size(template, &num_proc_rows);
  error += PLA_Temp_nb          (template, &nb);
  error += PLA_Temp_comm_row_rank       (template, &mycol),
  error += PLA_Temp_comm_row_size(template, &num_proc_cols),
  error += PLA_Temp_zero_or_one (template, &zero_or_one),  
 
  nb_col = num_proc_rows * nb;
  error += PLA_Temp_local_compute (1,global_j, 
				  mycol,
				  nb_col,
				  num_proc_cols,
				  zero_or_one,
				  &myn, rank);
  return (error);
} 

int PLA_Temp_matrix_owner_rowcol (PLA_Template template,
				 int global_i, 
				 int global_j, 
				 int *currow, 
				 int *curcol)
{
  int error=0;
  error += PLA_Temp_matrix_owner_row (template, 
				     global_i, 
				     currow);
  error += PLA_Temp_matrix_owner_col (template, 
				     global_j, 
				     curcol);
  return (error);
} 



int PLA_Temp_matrix_first_row_block    (PLA_Template     template,
				       int             global_length,
				       int             global_align_row,
				       int *           owner_row,
				       int *           top_block_size)
{
  int             error=0;
  int             global_block_size;
  int             adjusted_block_size;
  PLA_Start_num    zero_or_one;

  error += PLA_Temp_zero_or_one(template, 
                               &zero_or_one);
  error += PLA_Temp_nb         (template,
                               &global_block_size);
    
  error += PLA_Temp_matrix_owner_row (template,
				     global_align_row, 
				     owner_row);
    /*
       See Comments in Vector Section
       */


/* the if loop below is added to take care of the case when the aligment
of the matrix is not a multiple of a block size
Aug 22 1996 Philip Alpatov */

  
  adjusted_block_size = 
    (global_align_row - (int)zero_or_one)%global_block_size == 0 
                             ? 
     min(global_block_size, global_length) 
	                     : 
     min(global_block_size - (global_align_row - (int)zero_or_one)%
	 global_block_size, global_length);


  error += PLA_Temp_matrix_distant_length (template,
					  *owner_row,
					  adjusted_block_size, 
					  global_align_row, 
					  top_block_size);
  return(error);
}



int PLA_Temp_vector_first_row_block   (PLA_Template     template,      
				      int             global_length,
				      int             global_align_row,
				      int *           owner_row,
				      int *           top_block_size)
{
  int             error=0;
  int             global_block_size;
  int             adjusted_block_size;
  PLA_Start_num    zero_or_one;

  error += PLA_Temp_zero_or_one(template, 
                               &zero_or_one);
  error += PLA_Temp_nb         (template,
                               &global_block_size);
    
  error += PLA_Temp_vector_owner_rank (template,
				      global_align_row,
				      owner_row);
  
    /* 
       The block size is the upper bound on how many elements will be 
       stored on the beginning processor before wrapping to the second.
       But it is possible that because of the offset (stored in template)
       that only a partial block got stored in the first processor and
       the rest of the matrix was stored in the second processor.  
       
       To check this, how much of a vector of length block_size starting
       at the offset specified in template will be stored at the processor
       specified by owner_row?

       NOTE: This can also be done with a simple mod statement.  
       But in the general case this will be much more difficult and I think
       I should push these calculations down as much as possible to make
       going to the second level easier.
       */

  
  adjusted_block_size = 
    (global_align_row - (int)zero_or_one)%global_block_size == 0 
                             ? 
     min(global_block_size, global_length) 
	                     : 
     min(global_block_size - (global_align_row - (int)zero_or_one)%global_block_size,
	 global_length);


  error += PLA_Temp_vector_distant_length  (template,
					   *owner_row,
					   adjusted_block_size,
					   global_align_row,
					   top_block_size);
  return(error);
}


int PLA_Temp_matrix_last_row_block    (PLA_Template template,
				      int         global_length,
				      int         global_align_row,
				      int *       owner_row,
				      int *       bottom_block_length)
{
  int             error=0;
  int             global_block_size;
  int             combined_length; 
  int             block_length;
  PLA_Start_num    zero_or_one;
  
  error += PLA_Temp_zero_or_one(template, 
                               &zero_or_one);
  error += PLA_Temp_nb         (template,
                               &global_block_size);

  /* 
     I am thinking that global_align_row already has the zero_or_one 
     taken into account, so that just adding the length of the matrix
     gives you the offset to the last row.
     */
  combined_length = global_length + global_align_row;
  /* Want an offset not an alignment so JAG 2/17/97 */
/*   if(combined_length > 0) combined_length--;  */
 
  error += PLA_Temp_matrix_owner_row (template,
		     combined_length>0?combined_length-1:combined_length, 
				     owner_row);

  /*
     So for a vector aligned at the beginning of the template we only
     have to do a mod by the blocksize to compute how much is in that
     last block of the vector.  So add the offset to get a vector that
     is aligned with the beginning of the template and do a mod.  If 
     the offset is 1 and the length is 1, then we should always get
     a length of 1.  So take out zero_or_one.
     If the offset is 2 and the length is 1, then we should get a length
     of 1, which means we should check if the block is less than the 
     length of the matrix.
     */
  
  combined_length  -= (int) zero_or_one;
  block_length      = combined_length % global_block_size;  
  if (block_length > global_length)  block_length = global_length;
  
/* baker, 8.29.96
 * error caused when combined length a multiple of block_size
 * if global_length > block_size then block_length of last row equal zero
 */

  if ( global_length > 0 && block_length == 0 ) block_length = 
    min(global_block_size, global_length); /* JAG */

  *bottom_block_length = block_length;
  return(error);
}

int PLA_Temp_vector_last_row_block    (PLA_Template template,
				      int         global_length,
				      int         global_align_row,
				      int *       owner_row,
				      int *       bottom_block_length)
{
  int             error=0;
  int             global_block_size;
  int             combined_length; 
  int             block_length;
  PLA_Start_num    zero_or_one;
  
  error += PLA_Temp_zero_or_one(template, 
                               &zero_or_one);
  error += PLA_Temp_nb         (template,
                               &global_block_size);

  /* 
     I am thinking that global_align_row already has the zero_or_one 
     taken into account, so that just adding the lenght of the matrix
     gives you the offset to the last row.
     */
  combined_length = global_length + global_align_row;
  /* 
     If global_length was 1, global_align_row 0, then we would want 
     to pass a 0 to PLA_Temp_vector_owner_rank.  Because 
     PLA_Temp_vector_owner_rank omputes the owner of a vector of
     alignment combined_length.
     */
  /* 
     Make sure not to change the variable combined_length, only
     the value passed to the subroutine.  We still need to 
     determine the length of the last block.
     */
  error += PLA_Temp_vector_owner_rank (template,
		combined_length>0?combined_length-1:combined_length,
				      owner_row);

  /*
     So for a vector aligned at the beginning of the template we only
     have to do a mod by the blocksize to compute how much is in that
     last block of the vector.  So add the offset to get a vector that
     is aligned with the beginning of the template and do a mod.
     */
  
  combined_length  -= (int) zero_or_one;
  block_length      = combined_length % global_block_size;  
  if (block_length > global_length)  block_length = global_length;
  
/* baker, 8.29.96
 * error caused when combined length a multiple of block_size
 * if global_length > block_size then block_length of last row equal zero
 */

  if ( global_length > 0 && block_length == 0 ) block_length = 
    min(global_block_size, global_length); /* JAG */

  *bottom_block_length = block_length;
  return(error);
}


int PLA_Temp_matrix_first_col_block   (PLA_Template   template,  
				      int           global_width,
				      int           global_align_col,
				      int *         owner_col,
				      int *         left_block_width)
{
  int             error=0;
  int             global_block_size;
  int             adjusted_block_size;
  PLA_Start_num    zero_or_one;

  error += PLA_Temp_zero_or_one(template, 
                               &zero_or_one);
  error += PLA_Temp_nb         (template,
                               &global_block_size);
    
  error += PLA_Temp_matrix_owner_col (template,
				     global_align_col, 
				     owner_col);
  
    /* 
       The block size is the upper bound on how many elements will be 
       stored on the beginning processor before wrapping to the second.
       But it is possible that because of the offset (stored in template)
       that only a partial block got stored in the first processor and
       the rest of the matrix was stored in the second processor.  
       
       To check this, how much of a vector of width block_size starting
       at the offset specified in template will be stored at the processor
       specified by owner_row?
       */

  



/* the if loop below is added to take care of the case when the aligment
of the matrix is not a multiple of a block size
Aug 22 1996 Philip Alpatov */
  
  adjusted_block_size = 
    (global_align_col - (int)zero_or_one)%global_block_size == 0 
                             ? 
     min(global_block_size, global_width) 
	                     : 
     min(global_block_size - (global_align_col - (int)zero_or_one)%global_block_size,
	 global_width);
	    

  error += PLA_Temp_matrix_distant_width (template,
					 *owner_col,
					 adjusted_block_size, 
					 global_align_col, 
					 left_block_width);
  return(error);
}



int PLA_Temp_matrix_last_col_block   (PLA_Template   template,   
				     int           global_width,
				     int           global_align_col,
				     int *         owner_col,
				     int *         right_block_width)
{
  int             error=0;
  int             global_block_size;
  int             combined_width; 
  int             block_width;
  PLA_Start_num    zero_or_one;

  error += PLA_Temp_zero_or_one(template, 
                               &zero_or_one);
  error += PLA_Temp_nb         (template,
                               &global_block_size);
   
  /* 
     See comments in last_row_block.
     */

  combined_width = global_width + global_align_col;
  /* Want an offset not an alignment so JAG 2/17/97 */
/*  if(combined_width > 0) combined_width--;  */
  error += PLA_Temp_matrix_owner_col (template,
		     combined_width>0?combined_width-1:combined_width, 
				     owner_col);
  /*
     So for a vector aligned at the beginning of the template we only
     have to do a mod by the blocksize to compute how much is in that
     last block of the vector.  So add the offset to get a vector that
     is aligned with the beginning of the template and do a mod.
     */

  combined_width   -= (int) zero_or_one;
  block_width       = combined_width % global_block_size;  
  if (block_width > global_width)  block_width = global_width;
    

  if ( global_width > 0 && block_width == 0 ) block_width = 
    min(global_block_size, global_width); /* JAG */

  *right_block_width = block_width;
  return(error);
}



int PLA_Temp_matrix_first_row_first_col_block  (PLA_Template  template,
					       int          global_length,
					       int          global_width,
					       int          global_align_row,
					       int          global_align_col,
					       int *        owner_row,
					       int *        owner_col,
					       int *        topleft_block_size)
{
  int             error=0;
  int             top_length, left_width;

  error += PLA_Temp_matrix_first_row_block   (template,
					     global_length,
					     global_align_row,
					     owner_row,
					     &top_length);
  error += PLA_Temp_matrix_first_col_block   (template,
					     global_width,
					     global_align_col,
					     owner_col,
					     &left_width);
  *topleft_block_size =
    (top_length < left_width) ? top_length : left_width;
  
  return(error);
}



int PLA_Temp_matrix_last_row_last_col_block  (PLA_Template  template, 
					     int          global_length,
					     int          global_width,
					     int          global_align_row,
					     int          global_align_col,
					     int *        owner_row,
					     int *        owner_col,
					     int *        bottomright_block_size)
{
  int             error=0;
  int             bottom_length, right_width;

  error += PLA_Temp_matrix_last_row_block   (template,
					    global_length,
					    global_align_row,
					    owner_row,
					    &bottom_length);
  error += PLA_Temp_matrix_last_col_block   (template,
					    global_width,
					    global_align_col,
					    owner_col,
					    &right_width);
  *bottomright_block_size =
    (bottom_length < right_width) ? bottom_length : right_width;
  
  return(error);
}


int PLA_Temp_matrix_first_row_last_col_block  (PLA_Template  template,  
					      int          global_length,
					      int          global_width,
					      int          global_align_row,
					      int          global_align_col,
					      int *        owner_row,
					      int *        owner_col,
					      int *        topright_block_size)
{
  int             error=0;
  int             top_length, right_width;

  error += PLA_Temp_matrix_first_row_block   (template,
					     global_length,
					     global_align_row,
					     owner_row,
					     &top_length);
  error += PLA_Temp_matrix_last_col_block    (template,
					     global_width,
					     global_align_col,
					     owner_col,
					     &right_width);
  *topright_block_size =
    (top_length < right_width) ? top_length : right_width;
  
  return(error);
}


int PLA_Temp_matrix_last_row_first_col_block  (PLA_Template  template,  
					      int          global_length,
					      int          global_width,
					      int          global_align_row,
					      int          global_align_col,
					      int *        owner_row,
					      int *        owner_col,
					      int *        bottomleft_block_size)
{
  int             error=0;
  int             bottom_length, left_width;

  error += PLA_Temp_matrix_last_row_block   (template,
					    global_length,
					    global_align_row,
					    owner_row,
					    &bottom_length);
  error += PLA_Temp_matrix_first_col_block  (template,
					    global_width,
					    global_align_col,
					    owner_col,
					    &left_width);
  *bottomleft_block_size =
    (bottom_length < left_width) ? bottom_length : left_width;
  
  return(error);
}





int PLA_Temp_set_comm_dir(PLA_Template    template,
                          int             direction)
{
   int return_value = 0;

   printf("Pipelined communication is not yet supported in the template\n");
   printf("as noted in your call to PLA_Temp_set_comm_dir.             \n");
   printf
    ("If you see this message -- please send mail to gunnels@cs.utexas.edu\n");
   return_value += UNIMPLIMENTED_FUNCTION;

   return(return_value);
}

int PLA_Temp_get_comm_dir(PLA_Template    template,
                          int            *direction)
{
   int return_value = 0;

   printf("Pipelined communication is not yet supported in the template\n");
   printf("as noted in your call to PLA_Temp_get_comm_dir.             \n");
   printf
     ("If you see this message -- please send mail to gunnels@cs.utexas.edu\n");
   return_value += UNIMPLIMENTED_FUNCTION;

   return(return_value);
}




