#include "PLA.h"
#include "POOCLA.h"

void PLA_test_ooc_dsyrk ( int my_pe, PLA_Template template )
{
  int length, width, ldim;
  double * a, * c, time, mflops;

  PLA_Obj a_obj=NULL, c_obj=NULL, alpha_obj = NULL, beta_obj = NULL; 
  POOCLA_Obj a_ooc=NULL, c_ooc=NULL;
  MPI_Comm comm1d;
  int      i, j, n, k, np, int_uplo, int_trans;
  double alpha = 1.0e+0, beta = -1.0e+0;

  PLA_Temp_comm_all ( template, & comm1d );
  MPI_Comm_size( comm1d, &np );

  POOCLA_activate_instrumentation ();

  if ( my_pe == 0 )
  {
     printf ( "n k\n" );
     scanf ( "%d%d", & n, & k );
  }

  MPI_Bcast ( & n, 1, MPI_INT, 0, comm1d );
  MPI_Bcast ( & k, 1, MPI_INT, 0, comm1d );

  a = pla_calloc ( n * k, sizeof ( double ) );
  c = pla_calloc ( n * n, sizeof ( double ) );

  if ( my_pe == 0 )
  {
     PLA_test_dgenmatrix ( "G", 1.0e+0, n, k, a, n );
     PLA_test_dgenmatrix ( "G", 1.0e+0, n, n, c, n );
  }

  MPI_Bcast ( a, n * k, MPI_DOUBLE, 0, comm1d );
  MPI_Bcast ( c, n * n, MPI_DOUBLE, 0, comm1d );


/*
 * create in-core and ooc matrices
 */
  POOCLA_create ( template, MPI_DOUBLE, n, k, 0, 0,
                   PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, "Mat_A", & a_ooc );
  POOCLA_create_PLA_Obj_conf_to ( a_ooc, PLA_MATRIX, & a_obj );
  PLA_Obj_set_to_minus_one ( a_obj );
  PLA_Test_gdist_to_local ( a, n, a_obj );
  POOCLA_write ( a_ooc, a_obj );
  POOCLA_validate_filestatus ( a_ooc );

  POOCLA_create ( template, MPI_DOUBLE, n, n, 0, 0,
                   PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, "Mat_C", & c_ooc );
  POOCLA_create_PLA_Obj_conf_to ( c_ooc, PLA_MATRIX, & c_obj );
  PLA_Obj_set_to_minus_one ( c_obj );
  PLA_Test_gdist_to_local ( c, n, c_obj );
  POOCLA_write ( c_ooc, c_obj );
  POOCLA_validate_filestatus ( c_ooc );

/*
 * initialize scaling parameters and constants
 */
  PLA_Mscalar_create(MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,&alpha_obj);
  PLA_Obj_set_to_one(alpha_obj);
  PLA_Mscalar_create(MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template,&beta_obj);
  PLA_Obj_set_to_minus_one(beta_obj);

  int_uplo =  PLA_LOWER_TRIANGULAR;
  int_trans = PLA_NO_TRANS;

/*
 * calculate serial syrk and keep result in c
 */
  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime();

  PLA_dsyrk ( "L", "N", & n, & k, & alpha, a, & n, & beta, c, & n );

  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime() - time;
  mflops = n * n * k / time * 1.0e-6 / np;

  if ( my_pe == 0 ) printf ( "Serial : %f seconds at %f mflops\n", time, mflops );

/*
 * calculate incore syrk and keep result in c_obj
 */
  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime();

  PLA_Syrk ( int_uplo, int_trans, alpha_obj, a_obj, beta_obj, c_obj );

  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime() - time;
  mflops = n * n * k / time * 1.0e-6 / np;

  if ( my_pe == 0 ) printf ( "Incore : %f seconds at %f mflops\n", time, mflops );

/*
 * calculation ooc syrk, results on disk
 */
  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime();

  POOCLA_Syrk ( int_uplo, int_trans, alpha_obj, a_ooc, beta_obj, c_ooc );

  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime() - time;
  mflops = n * n * k / time * 1.0e-6 / np;

  if ( my_pe == 0 ) printf ( "Out-of-Core : %f seconds at %f mflops\n", time, mflops );

  POOCLA_read_time ( & time );
  MPI_Reduce ( & time, & mflops, 1, MPI_DOUBLE, MPI_SUM, 0, comm1d );
  if ( my_pe == 0 ) printf ( "Average Read Time : %f seconds\n", mflops / np );

  POOCLA_write_time ( & time );
  MPI_Reduce ( & time, & mflops, 1, MPI_DOUBLE, MPI_SUM, 0, comm1d );
  if ( my_pe == 0 ) printf ( "Average Write Time : %f seconds\n", mflops / np );

/*
 * read ooc solution into a_obj
 */
  POOCLA_read ( c_ooc, PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, & a_obj );

/*  PLA_test_compare ( c, n, c_obj );*/
  PLA_test_compare ( c, n, a_obj );


  PLA_Obj_free ( & a_obj );
  PLA_Obj_free ( & c_obj);
  PLA_Obj_free ( & alpha_obj );
  PLA_Obj_free ( & beta_obj );
  POOCLA_free ( & a_ooc );
  POOCLA_free ( & c_ooc );
  pla_free ( a );
  pla_free ( c );
}
