#ifndef _POOCLA_H
#define _POOCLA_H

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include "mmpi.h"
#include "PLA.h"

#ifndef PLA_UNDEFINED
#define PLA_UNDEFINED NULL
#endif

#ifndef PLA_UNDEF
#define PLA_UNDEF PLA_UNDEFINED
#endif

#ifndef PLA_FAILURE
#define PLA_FAILURE -1
#endif

typedef struct POOCLA_OBJECT_STRUCT
{
   PLA_Template  template;
   MPI_Datatype  datatype;
   int           filestatus,
                 global_length,
                 global_width,
                 global_align_row,
                 global_align_col,
                 universal_length,
                 universal_width,
                 universal_align_row,
                 universal_align_col,
                 objtype,
                 proj_onto,
                 owner,
                 file_ldim,
               * number_views;
   char          basefilename [ 128 ],
                 datafilename [ 128 ];
   FILE *        datafile;
}
POOCLA_OBJECT_STRUCT;

typedef POOCLA_OBJECT_STRUCT * POOCLA_Obj;

int POOCLA_set_nb_incore ( int nb_row_incore, int nb_col_incore );
int POOCLA_set_nb_row_incore ( int nb_row_incore );
int POOCLA_set_nb_col_incore ( int nb_col_incore );

int POOCLA_get_nb_incore ( int * nb_row_incore, int * nb_col_incore );
int POOCLA_get_nb_row_incore ( int * nb_row_incore );
int POOCLA_get_nb_col_incore ( int * nb_col_incore );

int POOCLA_create ( PLA_Template   template,
                    MPI_Datatype   datatype,
                    int            universal_length,
                    int            universal_width,
                    int            universal_align_row,
                    int            universal_align_col,
                    int            objtype,
                    int            proj_onto,
                    int            owner,
                    char *         basefilename,
                    POOCLA_Obj *   ooc_object );

int POOCLA_free   ( POOCLA_Obj *   ooc_object );

int POOCLA_create_PLA_Obj_conf_to ( POOCLA_Obj ooc, int objtype, PLA_Obj * obj );

int POOCLA_view ( POOCLA_Obj ooc,
                  int     length,
                  int     width,  
                  int     align_row,
                  int     align_col,  
                  POOCLA_Obj * new_ooc );

int POOCLA_view_all     ( POOCLA_Obj   ooc_object,
                          POOCLA_Obj * ooc_view );

int POOCLA_split_size   ( POOCLA_Obj ooc_object,
                          int side,
                          int * size );

int POOCLA_view_shift   ( POOCLA_Obj ooc_object,
                          int length_top,
                          int width_left,
                          int width_right,
                          int length_bottom );

int POOCLA_horz_split_2 ( POOCLA_Obj ooc_object,
                          int size,
                          POOCLA_Obj * top_object,
                          POOCLA_Obj * bottom_object );

int POOCLA_vert_split_2 ( POOCLA_Obj ooc_object,
                          int size,
                          POOCLA_Obj * left_object,
                          POOCLA_Obj * right_object );

int POOCLA_split_4      ( POOCLA_Obj ooc_object,
                          int length,
                          int width,
                          POOCLA_Obj * topleft_object,
                          POOCLA_Obj * topright_object,
                          POOCLA_Obj * bottomleft_object,
                          POOCLA_Obj * bottomright_object );

int POOCLA_read   ( POOCLA_Obj        ooc_object,
                    int            incore_objtype,
                    int            incore_proj_onto,
                    int            incore_owner,
                    PLA_Obj *      incore_object );

int poocla_read_same_type ( POOCLA_Obj ooc, PLA_Obj obj );
int poocla_read_pmvector_to_matrix ( POOCLA_Obj ooc, PLA_Obj obj );

int POOCLA_write  ( POOCLA_Obj        ooc_object,
                    PLA_Obj        incore_object );
int poocla_write_same_type ( POOCLA_Obj ooc, PLA_Obj obj );
int poocla_write_matrix_to_pmvector ( POOCLA_Obj ooc, PLA_Obj obj );

int poocla_read_data   ( int length,
                         int width,
                         int typesize,
                         FILE * source,
                         int source_ldim,
                         void * destin,
                         int ldim_destin );

int poocla_write_data  ( int length,
                         int width,
                         int typesize,
                         void * source,
                         int source_ldim,
                         FILE * destin,
                         int ldim_destin );

int POOCLA_template            ( POOCLA_Obj, PLA_Template * );
int POOCLA_datatype            ( POOCLA_Obj, MPI_Datatype * );
int POOCLA_filestatus          ( POOCLA_Obj, int * );
int POOCLA_global_length    ( POOCLA_Obj, int * );
int POOCLA_global_width     ( POOCLA_Obj, int * );
int POOCLA_global_align_row ( POOCLA_Obj, int * );
int POOCLA_global_align_col ( POOCLA_Obj, int * );
int poocla_universal_length    ( POOCLA_Obj, int * );
int poocla_universal_width     ( POOCLA_Obj, int * );
int poocla_universal_align_row ( POOCLA_Obj, int * );
int poocla_universal_align_col ( POOCLA_Obj, int * );

int POOCLA_objtype   ( POOCLA_Obj, int * );
int POOCLA_proj_onto ( POOCLA_Obj, int * );
int POOCLA_owner     ( POOCLA_Obj, int * );
int POOCLA_file_ldim           ( POOCLA_Obj, int * );
int POOCLA_basefilename        ( POOCLA_Obj, char ** );
int POOCLA_datafilename        ( POOCLA_Obj, char ** );
int POOCLA_datafile            ( POOCLA_Obj, FILE ** );

int POOCLA_validate_filestatus ( POOCLA_Obj );

int poocla_number_views ( POOCLA_Obj, int * );
int poocla_increment_number_views ( POOCLA_Obj );
int poocla_decrement_number_views ( POOCLA_Obj );

int POOCLA_activate_instrumentation ();
int POOCLA_deactivate_instrumentation ();
int POOCLA_set_start_time ();
int POOCLA_increment_read_time ();
int POOCLA_increment_write_time ();
int POOCLA_reset_read_time ();
int POOCLA_reset_write_time ();
int POOCLA_read_time ( double * );
int POOCLA_write_time ( double * );

int POOCLA_Syrk ( int uplo, int trans, PLA_Obj alpha, POOCLA_Obj a,
                                        PLA_Obj beta, POOCLA_Obj c );
int POOCLA_Syrk_lt ( PLA_Obj alpha, POOCLA_Obj a, PLA_Obj beta, POOCLA_Obj c );

int POOCLA_Chol ( int uplo, POOCLA_Obj a );
int POOCLA_Chol_l ( POOCLA_Obj a );

int POOCLA_Trsm ( int side, int uplo, int transa, int diag,
                   PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b );
int POOCLA_Trsm_lun ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b );
int POOCLA_Trsm_lut ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b );
int POOCLA_Trsm_lln ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b );
int POOCLA_Trsm_llt ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b );
int POOCLA_Trsm_run ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b );
int POOCLA_Trsm_rut ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b );
int POOCLA_Trsm_rln ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b );
int POOCLA_Trsm_rlt ( int diag, PLA_Obj alpha, POOCLA_Obj a, POOCLA_Obj b );

/* checkpointing routines */


typedef struct POOCLA_CHECKPOINT_STRUCT
{
   int level;
   char file [ 32 ];
   int  line;
   char comment [ 32 ];
   void * prev_chkpt,
        * next_chkpt;
} POOCLA_CHECKPOINT_STRUCT;

typedef POOCLA_CHECKPOINT_STRUCT * POOCLA_Chkpt;

int POOCLA_activate_checkpointing ( char * filename );
int POOCLA_deactivate_checkpointing ( void );
int poocla_set_checkpoint ( const char * const file,
                            const int line,
                            char * comment );
int POOCLA_end_checkpoint ( void );

#define POOCLA_set_checkpoint(                     comment ) \
        poocla_set_checkpoint( __FILE__, __LINE__, comment )


/* utility files convenient to keep with ooc library */

int convert_datafile_to_ooc_matrix ( char * filename, int universal_length,
                                     int universal_width, MPI_Datatype datatype,
                                     PLA_Template template, char * oocname );
int initialize_ooc_identity_matrix ( POOCLA_Obj ooc );
int extract_diagonal_from_ooc_matrix ( POOCLA_Obj ooc, double *, int );

#endif /* not defined _POOCLA_H */
