#include "POOCLA.h"

int convert_datafile_to_ooc_matrix ( char * filename,
                                     int universal_length, int universal_width,
                                     MPI_Datatype datatype, PLA_Template template,
                                     char * oocname )
{
   int top_size, left_size, base, extra, offset, my_col, my_row, c_dim, r_dim,
       global_length, global_width, global_align_row, global_align_col,
       local_length, local_width, local_align_row, local_align_col;
   double * buffer, alpha=1.0e+0;
   MPI_Aint typesize;
   PLA_Obj obj = NULL;
   POOCLA_Obj ooc = NULL, top = NULL, bottom = NULL, right = NULL;
   FILE * pfile = NULL;

   PLA_Temp_comm_row_rank ( template, & my_col );
   PLA_Temp_comm_row_size ( template, & c_dim );
   PLA_Temp_comm_col_rank ( template, & my_row );
   PLA_Temp_comm_col_size ( template, & r_dim );

   MPI_Type_extent ( datatype, & typesize );
   pfile = fopen ( filename, "rb" );

   PLA_API_begin ();

   POOCLA_create ( template, datatype, universal_length, universal_width, 0, 0,
                    PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, oocname, & ooc );

   POOCLA_view_all ( ooc, & right );

   POOCLA_split_size ( right, PLA_SIDE_LEFT, & left_size );

   while ( left_size > 0 )
   {
      POOCLA_vert_split_2 ( right, left_size, & bottom, & right );
      POOCLA_split_size ( bottom, PLA_SIDE_TOP, & top_size );

      while ( top_size > 0 )
      {
         POOCLA_horz_split_2 ( bottom, top_size, & top, & bottom );

         POOCLA_global_length ( top, & global_length );
         POOCLA_global_width ( top, & global_width );
         POOCLA_global_align_row ( top, & global_align_row );
         POOCLA_global_align_col ( top, & global_align_col );

         PLA_Matrix_create ( datatype, global_length, global_width, template,
                             global_align_row, global_align_col, & obj );
         PLA_Obj_set_to_minus_one ( obj );

         PLA_Obj_API_open ( obj );

         base = global_length / r_dim;
         extra = global_length % r_dim;
         local_length = base + ( extra > my_row ? 1 : 0 );
         local_align_row = my_row * base + ( extra > my_row ? my_row : extra );

         base = global_width / c_dim;
         extra = global_width % c_dim;
         local_width = base + ( extra > my_col ? 1 : 0 );
         local_align_col = my_col * base + ( extra > my_col ? my_col : extra );

         offset = ( ( global_align_col + local_align_col ) * universal_length +
                  global_align_row + local_align_row ) * typesize;

         if ( local_length * local_width > 0 )
         {
            buffer = pla_calloc ( local_length * local_width, typesize );

            fseek ( pfile, offset, SEEK_SET );
            poocla_read_data ( local_length, local_width, typesize,
                               pfile, universal_length,
                               buffer, local_length );

            PLA_API_axpy_matrix_to_global ( local_length, local_width, & alpha,
                                            buffer, local_length, obj,
                                            local_align_row, local_align_col );
            pla_free ( buffer ); buffer = NULL;
         }

         PLA_Obj_API_close ( obj );

         POOCLA_write ( top, obj );

         POOCLA_split_size ( bottom, PLA_SIDE_TOP, & top_size );
      }

      POOCLA_split_size ( right, PLA_SIDE_LEFT, & left_size );
   }

   PLA_API_end ();

   POOCLA_free ( & top );
   POOCLA_free ( & bottom );
   POOCLA_free ( & right );
   POOCLA_free ( & ooc );

   PLA_Obj_free ( & obj );

   return ( 0 );
}

int initialize_ooc_identity_matrix ( POOCLA_Obj ooc )
{
   int size, offset, top_size, left_size, my_pe, length, width;
   double one = 1.0e+0;
   PLA_Template template = NULL;
   PLA_Obj obj = NULL, subobj = NULL;
   POOCLA_Obj diag = NULL, curdiag = NULL;

   POOCLA_template ( ooc, & template );
   PLA_Temp_comm_all_rank ( template, & my_pe );

   PLA_API_begin ();

   POOCLA_view_all ( ooc, & diag );

   POOCLA_split_size ( diag, PLA_SIDE_LEFT, & left_size );
   POOCLA_split_size ( diag, PLA_SIDE_TOP, & top_size );
   POOCLA_global_length ( diag, & length );
   POOCLA_global_width ( diag, & width );

   while ( ( size = min ( top_size, left_size ) ) > 0 )
   {
      POOCLA_view_all ( diag, & curdiag );
      POOCLA_view_shift ( diag, size, size, 0, 0 );
      POOCLA_view_shift ( curdiag, 0, 0, size - width, size - length );

      length -= size;
      width -= size;

      POOCLA_create_PLA_Obj_conf_to ( curdiag, PLA_MATRIX, & obj );
      PLA_Obj_set_to_zero ( obj );
      PLA_Obj_view_all ( obj, & subobj );

      offset = 0;
      PLA_Obj_API_open ( subobj );

      while ( offset < size )
      {
         if ( my_pe == 0 )
            PLA_API_axpy_matrix_to_global ( 1, 1, & one, & one, 1,
                                            subobj, offset, offset );
         offset++;
      }

      PLA_Obj_API_close ( subobj );


      POOCLA_write ( curdiag, obj );
      PLA_Obj_free ( & subobj );
      PLA_Obj_free ( & obj );

      POOCLA_split_size ( diag, PLA_SIDE_LEFT, & left_size );
      POOCLA_split_size ( diag, PLA_SIDE_TOP, & top_size );
   }

   PLA_API_end ();

   POOCLA_validate_filestatus ( ooc );

   POOCLA_free ( & diag );
   POOCLA_free ( & curdiag );
   PLA_Obj_free ( & obj );
   PLA_Obj_free ( & subobj );

   return ( 0 );
}


int extract_diagonal_from_ooc_matrix ( POOCLA_Obj ooc, double * diagbuf, int diag_length )
{
   int size, top_size, left_size, my_pe, length, width, count, offset;
   double alpha = 1.0e+0;
   PLA_Template template = NULL;
   PLA_Obj obj = NULL;
   POOCLA_Obj diag = NULL, curdiag = NULL;
   MPI_Comm comm = MPI_COMM_NULL;

   POOCLA_template ( ooc, & template );
   PLA_Temp_comm_all_rank ( template, & my_pe );
   PLA_Temp_comm_all ( template, & comm );

   PLA_API_begin ();

   count = 0;

   POOCLA_view_all ( ooc, & diag );
   POOCLA_split_size ( diag, PLA_SIDE_LEFT, & left_size );
   POOCLA_split_size ( diag, PLA_SIDE_TOP, & top_size );
   POOCLA_global_length ( diag, & length );
   POOCLA_global_width ( diag, & width );

   while ( ( size = min3 ( top_size, left_size, diag_length ) ) > 0 )
   {
      POOCLA_view_all ( diag, & curdiag );
      POOCLA_view_shift ( diag, size, size, 0, 0 );
      POOCLA_view_shift ( curdiag, 0, 0, size - width, size - length );

      length -= size;
      width -= size;
      diag_length -= size;

      POOCLA_read ( curdiag, PLA_MATRIX, PLA_UNDEF, PLA_UNDEF, & obj );

      offset = 0;
      PLA_Obj_API_open ( obj );

      while ( offset < size )
      {
         if ( my_pe == 0 )
         {
            PLA_API_axpy_global_to_matrix ( 1, 1, & alpha, obj, offset, offset,
                                            & diagbuf [ count ], 1 );
            count++;
         }
         PLA_Obj_API_sync ( obj );
         offset++;
      }

      PLA_Obj_API_close ( obj );
      PLA_Obj_free ( & obj );

      POOCLA_split_size ( diag, PLA_SIDE_LEFT, & left_size );
      POOCLA_split_size ( diag, PLA_SIDE_TOP, & top_size );
   }

   PLA_API_end ();

   MPI_Bcast ( diagbuf, diag_length, MPI_DOUBLE, 0, comm );

   POOCLA_free ( & diag );
   POOCLA_free ( & curdiag );
   PLA_Obj_free ( & obj );

   return ( 0 );
}
