#include "POOCLA.h"

int POOCLA_Syrk ( int uplo, int trans, PLA_Obj alpha, POOCLA_Obj a,
                                        PLA_Obj beta, POOCLA_Obj c )
{
   POOCLA_Syrk_lt ( alpha, a, beta, c );
   return ( 0 );
}

int POOCLA_Syrk_lt ( PLA_Obj alpha, POOCLA_Obj a, PLA_Obj beta, POOCLA_Obj c )
{
   int error, size, top_size, left_size, bottom_size;
   int nb;
   PLA_Obj Acur = NULL, Bcur = NULL, Ccur = NULL, xbeta = NULL;
   POOCLA_Obj a_right = NULL, a_col = NULL, a_top = NULL, a_bottom = NULL,
           c_right = NULL, c_col = NULL, c_top = NULL, c_bottom = NULL,
           dummy   = NULL;

   {
      PLA_Template template;
      PLA_Obj_template ( alpha, & template );
      PLA_Environ_nb_alg ( PLA_OP_PAN_PAN, template, & nb );
   }

   PLA_Mscalar_create_conf_to ( beta, PLA_ALL_ROWS, PLA_ALL_COLS, & xbeta );
   PLA_Copy ( beta, xbeta );

   POOCLA_view_all ( a, & a_right );
   POOCLA_split_size ( a_right, PLA_SIDE_LEFT, & left_size );

   while ( left_size > 0 )
   {
      POOCLA_vert_split_2 ( a_right, left_size, & a_col, & a_right );
      POOCLA_view_all ( c, & c_right );

      POOCLA_split_size ( c_right, PLA_SIDE_TOP, & top_size );
      POOCLA_split_size ( c_right, PLA_SIDE_LEFT, & left_size );

      POOCLA_horz_split_2 ( a_col, min ( top_size, left_size ), & a_top, & a_col );

      error = POOCLA_read ( a_top, PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, & Acur );
      if ( error != 0 )
         PLA_Log_message ( "POOCLA_Syrk_lt", "Read of a_top returned error\n" );
      

      while ( ( size = min ( top_size, left_size ) ) > 0 )
      {
         POOCLA_split_4 ( c_right, size, size, & c_top, & dummy,
                                                & c_col, & c_right );

         error = POOCLA_read ( c_top, PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, & Ccur );
         if ( error != 0 )
            PLA_Log_message ( "POOCLA_Syrk_lt", "Read of c_top returned error\n" );

         error = PLA_Syrk ( PLA_LOW_TRIAN, PLA_NO_TRANS, alpha, Acur, xbeta, Ccur );
         if ( error != 0 )
            PLA_Log_message ( "POOCLA_Syrk_lt", "PLA_Syrk returned error\n" );

         error = POOCLA_write ( c_top, Ccur );
         if ( error != 0 )
            PLA_Log_message ( "POOCLA_Syrk_lt", "Write of c_top returned error\n" );

         POOCLA_view_all ( a_col, & a_top );
         POOCLA_view_all ( c_col, & c_top );

         POOCLA_split_size ( c_top, PLA_SIDE_BOTTOM, & bottom_size );

         while ( bottom_size > 0 )
         {
            POOCLA_horz_split_2 ( a_top, -bottom_size, & a_top, & a_bottom );
            POOCLA_horz_split_2 ( c_top, -bottom_size, & c_top, & c_bottom );

            error = POOCLA_read ( a_bottom,
                                   PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, & Bcur );
            if ( error != 0 )
               PLA_Log_message ( "POOCLA_Syrk_lt",
                                 "Read of a_bottom returned error\n" );
            error = POOCLA_read ( c_bottom,
                                   PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, & Ccur );
            if ( error != 0 )
               PLA_Log_message ( "POOCLA_Syrk_lt",
                                 "Read of c_bottom returned error\n" );

            error = PLA_Gemm ( PLA_NO_TRANS, PLA_TRANS, alpha, Bcur, Acur, xbeta, Ccur );
            if ( error != 0 )
               PLA_Log_message ( "POOCLA_Syrk_lt", "PLA_Gemm returned error\n" );

            error = POOCLA_write ( c_bottom, Ccur );
            if ( error != 0 )
               PLA_Log_message ( "POOCLA_Syrk_lt",
                                 "Write of c_bottom returned error\n" );

            POOCLA_split_size ( c_top, PLA_SIDE_BOTTOM, & bottom_size );
         }

         PLA_Obj_view_swap ( & Bcur, & Acur );

         POOCLA_view_shift ( a_col, size, 0, 0, 0 );

         POOCLA_split_size ( c_right, PLA_SIDE_TOP, & top_size );
         POOCLA_split_size ( c_right, PLA_SIDE_TOP, & left_size );
      }

      PLA_Obj_set_to_one ( xbeta );
      POOCLA_validate_filestatus ( c );
      POOCLA_split_size ( a_right, PLA_SIDE_LEFT, & left_size );
   }

   PLA_Obj_free ( & Acur );
   PLA_Obj_free ( & Bcur );
   PLA_Obj_free ( & Ccur );
   PLA_Obj_free ( & xbeta );

   POOCLA_free ( & a_right );
   POOCLA_free ( & a_col );
   POOCLA_free ( & a_top );
   POOCLA_free ( & a_bottom );
   POOCLA_free ( & c_right );
   POOCLA_free ( & c_col );
   POOCLA_free ( & c_top );
   POOCLA_free ( & c_bottom );
   POOCLA_free ( & dummy );

   return ( 0 );
}
