
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/

#ifndef PLAPACK_UTIL_H
#define PLAPACK_UTIL_H

/* prototypes for PLA_Alg_nb.c */

int PLA_Environ_nb_alg ( int, PLA_Template, int * );
int pla_Environ_set_nb_alg ( int, int );

/* prototypes for PLA_Error_tests.c */

/**************** START OF CORE ROUTINES **************************************/
void pla_Log_abort(
  const char * const who ,   /* function */
  const char * const file,
  const int          line,
  const char * const msg ,
  int error,
  int original_error );

int pla_Log_error(
   const char * const who ,
   const char * const file,
   const int          line,
   const char * const msg,
   int          this_error,
   int          original_error
   );

int pla_Log_message(
   const char * const who ,
   const char * const file,
   const int          line,
   const char * const msg
   );
/**************** END OF CORE ROUTINES ****************************************/
/*************  START "STANDARD" ERROR TESTS **********************************/
int pla_Obj_object_valid(
   PLA_Obj obj,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error);

int pla_Obj_objtype_valid(
   PLA_Obj object,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_objtype_equal(
   PLA_Obj obj,
   int object_type,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_datatype_valid(
   PLA_Obj object,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_datatype_equal(
   PLA_Obj object,
   MPI_Datatype datatype,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_template_equal(
   PLA_Obj object,
   PLA_Template template,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_length_equal (
   PLA_Obj object,
   int size,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_width_equal (
   PLA_Obj object,
   int size,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_length_equal (
   PLA_Obj object,
   int size,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_width_equal (
   PLA_Obj object,
   int size,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_length_length_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_length_width_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_width_length_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_width_width_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_length_length_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_length_width_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_width_length_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_width_width_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

/*************  END "STANDARD" ERROR TESTS ************************************/

/************* START ENCODED --> STRING CONVERSIONS  **************************/
int pla_code_string_objtype(
   int Object_type,
   char *object_name);

int pla_code_string_datatype(
   MPI_Datatype Datatype,
   char *data_name);

void pla_errorcode_string(
   int error_code,
   char *error_code_string);

/************* END ENCODED --> STRING CONVERSIONS  ****************************/

/**************** START FLAG INQUIRY ROUTINES *********************************/
int   PLA_Get_check_sequential (int *check_sequential );
int   PLA_Set_check_sequential (int sequential_set );
int   PLA_Get_instrument (int *instrument );
int   PLA_Set_instrument (int instrument_set );
int   PLA_Get_check_parameters ( int *check_parameter );
int   PLA_Set_check_parameters ( int check_set );
/**************** END FLAG INQUIRY ROUTINES ***********************************/

/**************** START TEMPORARY COPY ROUTINE ********************************/
/* Copy a matrix to an mscalar */

int PLA_Copy_temp(PLA_Obj the_matrix, PLA_Obj the_mscalar);

/**************** END TEMPORARY COPY ROUTINE ********************************/

/**************** START OF STACK STUFF ****************************************/
struct Stack_node_base {
   char *Message;
   struct Stack_node_base *previous_frame;
   struct Stack_node_base *next_frame;
};

typedef struct Stack_node_base PLA_Stack_node;
typedef PLA_Stack_node *PLA_Stack_frame_obj;


/* Some globals -- we need them global */
static PLA_Stack_frame_obj PLA_TOP_OF_MESSAGE_STACK = NULL;    /* could do these by sets */
static PLA_Stack_frame_obj PLA_BOTTOM_OF_MESSAGE_STACK = NULL; /* in PLA_Init()          */


int PLA_Stack_get_tos( PLA_Stack_frame_obj *top_of_stack);
int PLA_Stack_set_tos( PLA_Stack_frame_obj  top_of_stack);
int PLA_Stack_push_current(char *Stack_string);
int PLA_Stack_pop_current(char *Stack_string);

/**************** END OF STACK STUFF ******************************************/

/******************************************************************************/
/******************************************************************************/
/*   END OF HEADERS ***********************************************************/
/*   START OF CODE  ***********************************************************/
/******************************************************************************/
/******************************************************************************/


/**************** START OF CORE ROUTINES **************************************/
#define PLA_Log_abort(subroutine, message, error, original_error)             \
   pla_Log_abort (subroutine, __FILE__, __LINE__, message, error, original_error)

#define PLA_Log_error(subroutine, message, error, total_error)              \
   pla_Log_error (subroutine, __FILE__, __LINE__, message, error, total_error)

#define PLA_Log_message(subroutine, message)                                \
   pla_Log_message (subroutine, __FILE__, __LINE__, message)

/**************** END OF CORE ROUTINES ****************************************/

/*************  START "STANDARD" ERROR TESTS **********************************/
#define PLA_Obj_object_valid( object, message, error, original_error ) \
do       \
if(error = pla_Obj_object_valid(object, __FILE__, __LINE__, message, original_error))       \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
}  while (0)

#define PLA_Obj_objtype_valid( object, message, error, original_error )      \
do                                                                     \
if(error = pla_Obj_objtype_valid(object, __FILE__, __LINE__, message, original_error))           \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_objtype_equal( object, object_type, message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_objtype_equal(object, object_type, __FILE__, __LINE__, message, original_error))       \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_datatype_valid( object, message, error, original_error )      \
do                                                                     \
if(error = pla_Obj_datatype_valid(object, __FILE__, __LINE__, message, original_error))           \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)


#define PLA_Obj_datatype_equal( object, datatype, message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_datatype_equal(object, datatype, __FILE__, __LINE__, message, original_error))       \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_template_equal( object, template, message, error, original_error )  \
do                                                                     \
if(error = pla_Obj_template_equal(object, template, __FILE__, __LINE__, message, original_error))        \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_global_length_equal( obj1, size , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_length_equal( obj1, size , __FILE__, __LINE__, message, original_error ))     \
{                                                                              \
   if(0 == original_error) original_error = error;                             \
} while (0)

#define PLA_Obj_global_width_equal( obj1, size , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_width_equal( obj1, size , __FILE__, __LINE__, message, original_error ))     \
{                                                                              \
   if(0 == original_error) original_error = error;                             \
} while (0)

#define PLA_Obj_local_length_equal( obj1, size , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_length_equal( obj1, size , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while(0)

#define PLA_Obj_local_width_equal( obj1, size , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_width_equal( obj1, size , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_global_length_length_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_length_length_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_global_length_width_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_length_width_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_global_width_length_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_width_length_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_global_width_width_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_width_length_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))    \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_local_length_length_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_length_length_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_local_length_width_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_length_width_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_local_width_length_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_width_length_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_local_width_width_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_width_length_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))    \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
}while(0);
/*************  END "STANDARD" ERROR TESTS ************************************/


/************* START ENCODED --> STRING CONVERSIONS  **************************/
/************* END ENCODED --> STRING CONVERSIONS  ****************************/

/**************** START FLAG INQUIRY ROUTINES *********************************/

/**************** END FLAG INQUIRY ROUTINES ***********************************/

/**************** START TEMPORARY COPY ROUTINE ********************************/
/* Copy a matrix to an mscalar */

/**************** END TEMPORARY COPY ROUTINE **********************************/


/**************** START STACK STUFF  ******************************************/
/**************** END OF STACK STUFF *****************************************/



/* prototypes for PLA_Fill_laobj.c */

int PLA_Fill_laobj( PLA_Obj laobj);
int PLA_Print_laobj( PLA_Obj laobj);


/* prototypes for PLA_Show.c */

int PLA_Show(PLA_Obj laobj);


/* prototypes for PLA_Zero_loc_obj.c */

int PLA_Zero_loc_obj(PLA_Obj laobj);


/* prototypes for PLA_Extract_matrix.c */

int PLA_Extract_matrix( PLA_Obj global_matrix, PLA_Obj local_matrix);


/* prototypes for PLA_set_contents.c */

int PLA_Obj_set_contents( PLA_Obj laobj, void *value_ptr);


/* prototypes for safe memory allocation */
  
void * pla_malloc ( int size );
void * pla_calloc ( int num, int size );
void pla_free ( void * ptr );
int pla_memory_alloc_count ( void );
int pla_memory_free_count ( void );

/* prototypes for PLA_random */

double PLA_random(int seed);


#endif /* PLAPACK_UTIL_H */
