#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

void PRISM_accel(PLA_Obj X) {
/*
   Acceleration technique for Beta iterartions.

   Take the range of [0,1] and stretch it to [-s,1+s] so that 
   1/2 stays fixed, where s = .3.  Want new range to be smaller 
   than [1/2 - sqrt(3)/2, sqrt(3)/2 + 1/2] since this is the 
   value where the beta function has the value of 1/2 (this 
   range is [-.366,1.366] which would be s = 0.366).

   Method:

     stretch = .3
     a = 1 + 2 * stretch
     b = -stretch
     X = aX + b
*/
    double temp;
    MPI_Datatype datatype;
    PLA_Template template;
    PLA_Obj alpha=NULL;
    const double stretch = .3;

    PLA_Obj_datatype( X, &datatype );
    PLA_Obj_template( X, &template );

    /* Temporary multiscalar */
    PLA_Mscalar_create(datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
        template, &alpha);

    /* compute X = aX + b */
    temp = 1. + 2. * stretch;
    PLA_Obj_set( alpha, datatype, &temp );
    PLA_Local_scal( alpha, X );

    temp = -stretch;
    PLA_Obj_set( alpha, datatype, &temp );
    PRISM_add_to_diag( alpha, X );

    PLA_Obj_free(&alpha);

#ifdef DB_GLOBAL_
    if (me==ROOT) printf(">> PRISM_accel\n");
#endif
}
