#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

void PRISM_beta(PLA_Obj X, PLA_Obj Beta, PLA_Obj temp) {
/*
    Apply beta function (3*X^2 - 2*X^3) to symmetric matrix X:

    temp =  X * X
    beta = -2 * X * temp 
    beta += 3 * temp
*/
    double value;
    PLA_Obj alpha=NULL;
    MPI_Datatype datatype;
    PLA_Template template;

    /* Extract info from matrix X */
    PLA_Obj_datatype( X, &datatype );
    PLA_Obj_template( X, &template );

    /* temporary scalar */
    PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS,
        1, 1, template, &alpha );

    PLA_Gemm( PLA_NO_TRANS, PLA_NO_TRANS, PRISM_one, X, X, PRISM_zero, temp );

    value = -2.;  
    PLA_Obj_set( alpha, datatype, &value );
    PLA_Gemm( PLA_NO_TRANS, PLA_NO_TRANS, alpha, X, temp, PRISM_zero, Beta );

    value = 3.;  
    PLA_Obj_set( alpha, datatype, &value );
    PLA_Local_axpy( alpha, temp, Beta );

    PLA_Obj_free( &alpha );
}


