/*******************************************************************
*  File: ns3d.c
* This application runs on sp2 parallel computer, whit using 
* MPI library.
*
*********************************************************************
*
* Modified by:
*             Adnan Agbaria
*
**********************************************************************/

#include "mpi.h"  /* The MPI header file. */
#include "ns3d.h"
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <errno.h>

#define max(x,y)      ((x)>(y)?(x):(y))
#define min(x,y)      ((x)<(y)?(x):(y))
#define MOD(x,y)      ((x)%(y))
#define NHOSTMAX  10


/**********************************************************
* To determine message passing interface!!
***********************************************************/
#define SEND_RECV   1
/*#define MSG         1*/
/*#define NONBLOCK    1*/

/* For process 1 */
#define GLOBAL_SR   1
/*#define GATHER      1*/


/**********************************************************************
|     simulate the parallel computation
|
|
|    dW/dt+NL=-\nabla P +ET \nabla^2 W,     W={U,V}
|
|       NL=rot W x W ,                      NL={G,H}
|
|      periodic b.c. in x and y
|
|         dP/dx = P_1(x,y) + PAV,
|           P_1 is periodic,
|   PAV - prescribed constant pressure gradient in x
|
*/

/*********************************************************************
* Global variables.
**********************************************************************/
int me, tids[20];              /* contain pid.*/
int buffer_size,position,nprocs, position2;   
char *buffer, *bufferSend, *bufferRecv; /*temp*/
int root;

MPI_Status status;

/**************************/
/* For P1 processor       */
/**************************/
long stim, utim;
float sTime, STime=0;


/**********************************************************************
* Function: main()
*
***********************************************************************/
void main(argc, argv)
int argc;
char *argv[];
{
  extern void strip(),transtrip(),transwec(),twcollect(),twdistr();

  int nn, num;
  
  static int K0 = 0;
  static int K1 = 1;
  static int K2 = 2;
  static int K3 = 3;
  static int K4 = 4;
  static int K5 = 5;
  static int K6 = 6;
  static int K7 = 7;
  
  auto int nstep;
  FILE *out_file;
  
  static float
    tur[KK2][MM2][NEP1+NEM+1], tul[KK2][MM2][NEP1+NEM+1],
    tvr[KK2][MM2][NEP1+NEM+1], tvl[KK2][MM2][NEP1+NEM+1],
    twr[KK2][MM2][NEP1+NEM+1], twl[KK2][MM2][NEP1+NEM+1],
    
    yur[KK2][MM2][NEP1+NEM+1], yul[KK2][MM2][NEP1+NEM+1],
    yvr[KK2][MM2][NEP1+NEM+1], yvl[KK2][MM2][NEP1+NEM+1],
    ywr[KK2][MM2][NEP1+NEM+1], ywl[KK2][MM2][NEP1+NEM+1],
/****/    
    tbul[KK2][MM2], tbur[KK2][MM2], tbvl[KK2][MM2],
    tbvr[KK2][MM2], tbwl[KK2][MM2], tbwr[KK2][MM2],
    
    tdul[KK2][MM2], tdur[KK2][MM2], tdvl[KK2][MM2],
    tdvr[KK2][MM2], tdwl[KK2][MM2], tdwr[KK2][MM2],
    
    ybul[KK2][MM2], ybur[KK2][MM2], ybvl[KK2][MM2],
    ybvr[KK2][MM2], ybwl[KK2][MM2], ybwr[KK2][MM2],
    
    ydul[KK2][MM2], ydur[KK2][MM2], ydvl[KK2][MM2],
    ydvr[KK2][MM2], ydwl[KK2][MM2], ydwr[KK2][MM2];
/******/  
  /* For the global communication. */
  static float
    ybpl[NPR][KK2][MM2],ybpr[NPR][KK2][MM2],ydpl[NPR][KK2][MM2],
    ydpr[NPR][KK2][MM2];
  
  static float work[NWORK],trigs[NTRIG];
  static int ifax[13],icfax[13];
  static float ywork[MWORK],ytrigs[MTRIG];
  static int iyfax[13];
  static float zwork[KWORK], ztrigs[KTRIG];
  static int izfax[13];
  static float yk1[MMYH],yk2[MMYH],zk1[MM2],zk2[MM2];
  static float wnz[KKZH],wnzd[KK2];
  static float xk1[NXH],xk2[NXH];
  static float b[NXH],s[NEP1+NEM+1],c[NEP1+NEM+1];
  static float par[25];
  static float dlr,dly;
  static float tn,tt,dt,al1,asq1,pav;
  static float am[NPR],ab[NPR];
  static int kx,ky,lx,ly;
  static int is;

  static float
    uu[KK2][MM2][NDM], vv[KK2][MM2][NDM], ww[KK2][MM2][NDM],
    uul[KK2][MM2][NDM], vvl[KK2][MM2][NDM], wwl[KK2][MM2][NDM],
    uus[KK2][MM2][NDM], vvs[KK2][MM2][NDM], wws[KK2][MM2][NDM],
    uup[KK2][MM2][NDM], vvp[KK2][MM2][NDM], wwp[KK2][MM2][NDM],

    ue[KK2][MM2][NDM], ve[KK2][MM2][NDM], we[KK2][MM2][NDM],
    vd[KK2][MM2][NDM], wd[KK2][MM2][NDM],
    
    fa[KK2][MM2][NDM], fb[KK2][MM2][NDM], fc[KK2][MM2][NDM],
    fd[KK2][MM2][NDM],
    ga[KK2][MM2][NDM], gb[KK2][MM2][NDM], gc[KK2][MM2][NDM],
    gd[KK2][MM2][NDM],
    ra[KK2][MM2][NDM], rb[KK2][MM2][NDM], rc[KK2][MM2][NDM],
    rd[KK2][MM2][NDM],
    gt[KK2][MM2][NDM], gm[KK2][MM2][NDM], gl[KK2][MM2][NDM],
    ht[KK2][MM2][NDM], hm[KK2][MM2][NDM], hl[KK2][MM2][NDM],
    lt[KK2][MM2][NDM], lm[KK2][MM2][NDM], ll[KK2][MM2][NDM];
  
  static int
    lchu[KK2][MM2],lchv[KK2][MM2],lchw[KK2][MM2];
  
  int i,j,k;
  
  int isize;
  int  nproc;
  int  left_id, right_id, msgtag = 5;
  
  /* Time variables.*/
  struct timeval begTime;
  struct tms begBuf;

  if( times( &begBuf ) < 0 )
    perror("times" );

  out_file = fopen("results.cc","a");
  if( gettimeofday(&begTime, (struct timezone*)0) < 0 )
    perror("gettimeofday");

  nstep = NSTEP;
  par[21]=nstep;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&me);
  MPI_Comm_size(MPI_COMM_WORLD,&nprocs);
  nproc = NPR;
  buffer_size =BUFFSIZE;
  
  buffer = malloc(BUFFSIZE);      /* temp!!!! */
  bufferSend = malloc(BUFFSIZE);  /* temp!!!! */
  bufferRecv = malloc(BUFFSIZE);  /* temp!!!! */

  /* so that we do not need to change much code */
  for (i=0; i < nproc; i++)
    tids[i] = i;     
  
  /* number of processor sometimes nessary to count from unity */
  nn = me + 1;

  /*------------------------------*/
  /*  initialization (NENTER=0)   */
  /*------------------------------*/
  strip(tur,tul,tvr,tvl,twr,twl,
	yur,yul,yvr,yvl,ywr,ywl,
	tbul,tbur,tdul,tdur,
	tbvl,tbvr,tdvl,tdvr,
	tbwl,tbwr,tdwl,tdwr,	
	ybpl,ybpr,ydpl,ydpr,
	ybul,ybur,ydul,ydur,
	ybvl,ybvr,ydvl,ydvr,
	ybwl,ybwr,ydwl,ydwr,
	&K0,out_file,nn,
	work,trigs,ywork,ytrigs,zwork,ztrigs,ifax,icfax,iyfax,izfax,
	am,ab,
	yk1,yk2,zk1,zk2,xk1,xk2,wnz,wnzd,b,s,c,par,
	uu,uul,uus,uup,ue,
	vv,vvl,vvs,vvp,ve,
	ww,wwl,wws,wwp,we,
	vd,wd,
	gt,gm,gl,
	ht,hm,hl,
	lt,lm,ll,
	fa,fb,fc,fd,
	ga,gb,gc,gd,
	ra,rb,rc,rd,		
	lchu,lchv,lchw,0,begTime,begBuf);

  /*------------------------------------------*/
  /* data exchange with neighbours  /L - /R   */
  /*------------------------------------------*/
  right_id = ( me != (nproc-1) ) ? tids[me+1] : tids[0];
  left_id = ( me != 0 ) ? tids[me-1] : tids[nproc-1];

  /* All the communications inside the loop.*/
  for(is=1; is<=nstep; is++){

    /*-------------------------------------*/
    /*  Explicit step          (NENTER=1)  */
    /*-------------------------------------*/
    strip(tur,tul,tvr,tvl,twr,twl,
	  yur,yul,yvr,yvl,ywr,ywl,
	  tbul,tbur,tdul,tdur,
	  tbvl,tbvr,tdvl,tdvr,
	  tbwl,tbwr,tdwl,tdwr,
	  ybpl,ybpr,ydpl,ydpr,
	  ybul,ybur,ydul,ydur,
	  ybvl,ybvr,ydvl,ydvr,
	  ybwl,ybwr,ydwl,ydwr,
	  &K1,out_file,nn,
	  work,trigs,ywork,ytrigs,zwork, ztrigs, ifax,icfax,iyfax, izfax,
	  am,ab,
	  yk1,yk2,zk1,zk2,xk1,xk2,wnz,wnzd,b,s,c,par,
	  uu,uul,uus,uup,ue,
	  vv,vvl,vvs,vvp,ve,
	  ww,wwl,wws,wwp,we,
	  vd,wd,
	  gt,gm,gl,
	  ht,hm,hl,
	  lt,lm,ll,
	  fa,fb,fc,fd,
	  ga,gb,gc,gd,
	  ra,rb,rc,rd,
	  lchu,lchv,lchw,is,begTime,begBuf);
    
    isize  = (KK2)*(MM2)*(NEP1+NEM+1);
    

/************************ SEND_RECV ***********************************/
    /*send message to the left neighbour + receive from right */
    MPI_Sendrecv(tul, isize, MPI_FLOAT, left_id, msgtag,
		 yur, isize, MPI_FLOAT, right_id, msgtag,
		 MPI_COMM_WORLD,&status);
    
    /*send message to the right neighbour + receive from left */
    MPI_Sendrecv(tur, isize, MPI_FLOAT, right_id, msgtag,
		 yul, isize, MPI_FLOAT, left_id, msgtag,
		 MPI_COMM_WORLD,&status);


    /*-------------------------------------------------*/
    /*  RHS in Poission equation. NENTER=2             */
    /*-------------------------------------------------*/
    strip(tur,tul,tvr,tvl,twr,twl,
	  yur,yul,yvr,yvl,ywr,ywl,
	  tbul,tbur,tdul,tdur,
	  tbvl,tbvr,tdvl,tdvr,
	  tbwl,tbwr,tdwl,tdwr,
	  ybpl,ybpr,ydpl,ydpr,
	  ybul,ybur,ydul,ydur,
	  ybvl,ybvr,ydvl,ydvr,
	  ybwl,ybwr,ydwl,ydwr,
	  &K2,out_file,nn,
	  work,trigs,ywork,ytrigs,zwork, ztrigs,ifax,icfax,iyfax,izfax,
	  am,ab,
	  yk1,yk2,zk1,zk2,xk1,xk2,wnz,wnzd,b,s,c,par,
	  uu,uul,uus,uup,ue,
	  vv,vvl,vvs,vvp,ve,
	  ww,wwl,wws,wwp,we,
	  vd,wd,
	  gt,gm,gl,
	  ht,hm,hl,
	  lt,lm,ll,
	  fa,fb,fc,fd,
	  ga,gb,gc,gd,
	  ra,rb,rc,rd,
	  lchu,lchv,lchw,is,begTime,begBuf);
    
    /*------------------------------------------*/
    /* data exchange with neighbours  /L - /R   */
    /*------------------------------------------*/

/********************** SEND_RECV ***********************************/
    /*send message to the left neighbour + receive from right */
    MPI_Sendrecv(tul, isize, MPI_FLOAT, left_id, msgtag,
		 yur, isize, MPI_FLOAT, right_id, msgtag,
		 MPI_COMM_WORLD,&status);
    
    /*send message to the right neighbour + receive from left */
    MPI_Sendrecv(tur, isize, MPI_FLOAT, right_id, msgtag,
		 yul, isize, MPI_FLOAT, left_id, msgtag,
		 MPI_COMM_WORLD,&status);


    /*-------------------------------------*/
    /* poission equation    (NENTER=3)     */
    /*-------------------------------------*/
    strip(tur,tul,tvr,tvl,twr,twl,
	  yur,yul,yvr,yvl,ywr,ywl,
	  tbul,tbur,tdul,tdur,
	  tbvl,tbvr,tdvl,tdvr,
	  tbwl,tbwr,tdwl,tdwr,
	  ybpl,ybpr,ydpl,ydpr,
	  ybul,ybur,ydul,ydur,
	  ybvl,ybvr,ydvl,ydvr,
	  ybwl,ybwr,ydwl,ydwr,
	  &K3,out_file,nn,
	  work,trigs,ywork,ytrigs,zwork, ztrigs, ifax,icfax,iyfax,izfax,
	  am,ab,
	  yk1,yk2,zk1,zk2,xk1,xk2,wnz,wnzd,b,s,c,par,
	  uu,uul,uus,uup,ue,
	  vv,vvl,vvs,vvp,ve,
	  ww,wwl,wws,wwp,we,
	  vd,wd,
	  gt,gm,gl,
	  ht,hm,hl,
	  lt,lm,ll,
	  fa,fb,fc,fd,
	  ga,gb,gc,gd,
	  ra,rb,rc,rd,
	  lchu,lchv,lchw,is,begTime,begBuf);
    
    /*
      |     G L O B A L data transfer:
      | _____________
      |             |
      |  1 <-- 2,   |     3 <-- 4,       5 <-- 6,       7 <-- 8
      |             |___________
      |                        |
      |           1 <-- 3,     |     5 <-- 7
      | _______________________|
      |
      |                     1 <-- 5
      */
    
    /*-------------------------------------*/
    /* all channels  -> 1                  */
    /*-------------------------------------*/
    isize = KK2*MM2;
    

/***************************** GLOBAL_SR **********************************/
    
    /* All the processors, send the data to P1.*/
    if( me != 0 ){
      position = 0;
      /* To send one packet to P1, is more quikly!!!*/ 
      MPI_Pack(tbul, isize, MPI_FLOAT, bufferSend, buffer_size,
	       &position, MPI_COMM_WORLD);
      MPI_Pack(tbur, isize, MPI_FLOAT, bufferSend, buffer_size,
	       &position, MPI_COMM_WORLD);
      MPI_Pack(tdul, isize, MPI_FLOAT, bufferSend, buffer_size,
	       &position,MPI_COMM_WORLD);
      MPI_Pack(tdur, isize, MPI_FLOAT, bufferSend, buffer_size,
	       &position,MPI_COMM_WORLD);
      MPI_Send(bufferSend, position, MPI_PACKED, tids[0], me, MPI_COMM_WORLD);
      

    } /* if (me != 0) */
    
    /* first channel (P1) receives data and compute the poission.*/
    else{
      for( i = 1; i < NPR; i++ ){
	position = 0;
	MPI_Recv(bufferRecv, buffer_size, MPI_PACKED, tids[i], i, 
		 MPI_COMM_WORLD, &status);
	/*printf("recv %d<-%d\n",me,tids[i]);*/
	MPI_Unpack(bufferRecv, buffer_size, &position, ybpl[i], isize, 
		   MPI_FLOAT, MPI_COMM_WORLD);
	MPI_Unpack(bufferRecv, buffer_size, &position, ybpr[i], isize, 
		   MPI_FLOAT, MPI_COMM_WORLD);
	MPI_Unpack(bufferRecv, buffer_size, &position, ydpl[i], isize, 
		   MPI_FLOAT, MPI_COMM_WORLD);
	MPI_Unpack(bufferRecv, buffer_size, &position, ydpr[i], isize, 
		   MPI_FLOAT, MPI_COMM_WORLD);
      }

      /*-------------------------------------*/
      /*  compute matching paramters in      */
      /* processor 1 (P1      (NENTER=4)     */
      /*-------------------------------------*/
      strip(tur,tul,tvr,tvl,twr,twl,
	    yur,yul,yvr,yvl,ywr,ywl,
	    tbul,tbur,tdul,tdur,
	    tbvl,tbvr,tdvl,tdvr,
	    tbwl,tbwr,tdwl,tdwr,
	    ybpl,ybpr,ydpl,ydpr,
	    ybul,ybur,ydul,ydur,
	    ybvl,ybvr,ydvl,ydvr,
	    ybwl,ybwr,ydwl,ydwr,
	    &K4,out_file,nn,
	    work,trigs,ywork,ytrigs,zwork,ztrigs,ifax,icfax,iyfax,izfax,
	    am,ab,
	    yk1,yk2,zk1,zk2,xk1,xk2,wnz,wnzd,b,s,c,par,
	    uu,uul,uus,uup,ue,
	    vv,vvl,vvs,vvp,ve,
	    ww,wwl,wws,wwp,we,
	    vd,wd,
	    gt,gm,gl,
	    ht,hm,hl,
	    lt,lm,ll,
	    fa,fb,fc,fd,
	    ga,gb,gc,gd,
	    ra,rb,rc,rd,
	    lchu,lchv,lchw,is,begTime,begBuf);

      /*-------------------------------------*/
      /* return matching coefficients        */
      /*-------------------------------------*/
      for( i = 1; i < NPR; i++ ){
	position = 0;
	MPI_Pack(ybpl[i], isize, MPI_FLOAT, bufferSend, buffer_size,
		 &position,MPI_COMM_WORLD);
	MPI_Pack(ybpr[i], isize, MPI_FLOAT, bufferSend, buffer_size,
		 &position,MPI_COMM_WORLD);
	MPI_Send(bufferSend, position, MPI_PACKED, tids[i], i, 
		 MPI_COMM_WORLD);

      }
    } /*else, processor 1.*/

    /*-------------------------------------*/
    /* receive matching coefficients       */
    /*-------------------------------------*/
    if( me != 0 ){
      position = 0;
      MPI_Recv(bufferRecv, buffer_size, MPI_PACKED, tids[0], me, 
	       MPI_COMM_WORLD, &status);
      MPI_Unpack(bufferRecv, buffer_size, &position, tbul, isize, 
		 MPI_FLOAT, MPI_COMM_WORLD);
      MPI_Unpack(bufferRecv, buffer_size, &position, tbur, isize, 
		 MPI_FLOAT, MPI_COMM_WORLD);
      
    }

    
    /*-------------------------------------*/
    /*  global matching in                 */
    /* poission equation    (NENTER=5)     */
    /*-------------------------------------*/
    strip(tur,tul,tvr,tvl,twr,twl,
	  yur,yul,yvr,yvl,ywr,ywl,
	  tbul,tbur,tdul,tdur,
	  tbvl,tbvr,tdvl,tdvr,
	  tbwl,tbwr,tdwl,tdwr,
	  ybpl,ybpr,ydpl,ydpr,
	  ybul,ybur,ydul,ydur,
	  ybvl,ybvr,ydvl,ydvr,
	  ybwl,ybwr,ydwl,ydwr,
	  &K5,out_file,nn,
	  work,trigs,ywork,ytrigs,zwork, ztrigs,ifax,icfax,iyfax,izfax,
	  am,ab,
	  yk1,yk2,zk1,zk2,xk1,xk2,wnz,wnzd,b,s,c,par,
	  uu,uul,uus,uup,ue,
	  vv,vvl,vvs,vvp,ve,
	  ww,wwl,wws,wwp,we,
	  vd,wd,
	  gt,gm,gl,
	  ht,hm,hl,
	  lt,lm,ll,
	  fa,fb,fc,fd,
	  ga,gb,gc,gd,
	  ra,rb,rc,rd,
	  lchu,lchv,lchw,is,begTime,begBuf);

    /*----------------------------------------*/
    /* data exchange with neighbours /L - /R  */
    /*----------------------------------------*/
    isize  = KK2*MM2*(NEP1+NEM+1);
    
/*************************** SEND_RECV ********************************/

    /*send message to the left neighbour + receive from right */
    position = 0;
    MPI_Pack(tul, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tvl, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(twl, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);

    MPI_Sendrecv(bufferSend, position, MPI_PACKED, left_id, msgtag, 
		 bufferRecv, buffer_size, MPI_PACKED, right_id, msgtag,
		 MPI_COMM_WORLD, &status);
    
    position = 0;
    MPI_Unpack(bufferRecv, buffer_size, &position, yur, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, yvr, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ywr, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    
    /*send message to the right neighbour + receive from left */
    position = 0;
    MPI_Pack(tur, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tvr, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(twr, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Sendrecv(bufferSend, position, MPI_PACKED, right_id, msgtag,
		 bufferRecv, buffer_size, MPI_PACKED, left_id, msgtag,
		 MPI_COMM_WORLD,&status);
    
    position = 0;
    MPI_Unpack(bufferRecv, buffer_size, &position, yul, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, yvl, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ywl, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    


    /*---------------------------------------*/
    /* particular solution UU1,VV1 ...)      */
    /* interface data (BUL,BUR1,..)(NENTER=6)*/
    /*---------------------------------------*/
    strip(tur,tul,tvr,tvl,twr,twl,
	  yur,yul,yvr,yvl,ywr,ywl,
	  tbul,tbur,tdul,tdur,
	  tbvl,tbvr,tdvl,tdvr,
	  tbwl,tbwr,tdwl,tdwr,
	  ybpl,ybpr,ydpl,ydpr,
	  ybul,ybur,ydul,ydur,
	  ybvl,ybvr,ydvl,ydvr,
	  ybwl,ybwr,ydwl,ydwr,
	  &K6,out_file,nn,
	  work,trigs,ywork,ytrigs,zwork, ztrigs,ifax,icfax,iyfax,izfax,
	  am,ab,
	  yk1,yk2,zk1,zk2,xk1,xk2,wnz,wnzd,b,s,c,par,
	  uu,uul,uus,uup,ue,
	  vv,vvl,vvs,vvp,ve,
	  ww,wwl,wws,wwp,we,
	  vd,wd,
	  gt,gm,gl,
	  ht,hm,hl,
	  lt,lm,ll,
	  fa,fb,fc,fd,
	  ga,gb,gc,gd,
	  ra,rb,rc,rd,
	  lchu,lchv,lchw,is,begTime,begBuf);
    
    /*-------------------------------------*/
    /* exchange of boundary values /L - /R */
    /*-------------------------------------*/
    /*send message to the left neighbour*/
    isize = KK2*MM2;


/************************** SEND_RECV *******************************/

    /* send to left + receive from right */
    position = 0;
    MPI_Pack(tbul, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tdul, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tbvl, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tdvl, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tbwl, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tdwl, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Sendrecv(bufferSend, position, MPI_PACKED, left_id, msgtag, 
		 bufferRecv, buffer_size, MPI_PACKED, right_id, msgtag,
		 MPI_COMM_WORLD,&status);
    
    position = 0;
    MPI_Unpack(bufferRecv, buffer_size, &position, ybur, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ydur, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ybvr, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ydvr, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ybwr, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ydwr, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    
    /*send message to the right neighbour + receive from left */
    position = 0;
    MPI_Pack(tbur, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tdur, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tbvr, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tdvr, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tbwr, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Pack(tdwr, isize, MPI_FLOAT, bufferSend, buffer_size,
	     &position,MPI_COMM_WORLD);
    MPI_Sendrecv(bufferSend, position, MPI_PACKED, right_id, msgtag, 
		 bufferRecv, buffer_size, MPI_PACKED, left_id, msgtag,
		 MPI_COMM_WORLD,&status);
    
    position = 0;
    MPI_Unpack(bufferRecv, buffer_size, &position, ybul, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ydul, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ybvl, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ydvl, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ybwl, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    MPI_Unpack(bufferRecv, buffer_size, &position, ydwl, isize, 
	       MPI_FLOAT, MPI_COMM_WORLD);
    

    /*---------------------------------------*/
    /* matching homogenous solution on n+1   */
    /* time step     NENTER=7                */
    /*---------------------------------------*/
    strip(tur,tul,tvr,tvl,twr,twl,
	  yur,yul,yvr,yvl,ywr,ywl,
	  tbul,tbur,tdul,tdur,
	  tbvl,tbvr,tdvl,tdvr,
	  tbwl,tbwr,tdwl,tdwr,
	  ybpl,ybpr,ydpl,ydpr,
	  ybul,ybur,ydul,ydur,
	  ybvl,ybvr,ydvl,ydvr,
	  ybwl,ybwr,ydwl,ydwr,
	  &K7,out_file,nn,
	  work,trigs,ywork,ytrigs,zwork, ztrigs,ifax,icfax,iyfax,izfax,
	  am,ab,
	  yk1,yk2,zk1,zk2,xk1,xk2,wnz,wnzd,b,s,c,par,
	  uu,uul,uus,uup,ue,
	  vv,vvl,vvs,vvp,ve,
	  ww,wwl,wws,wwp,we,
	  vd,wd,
	  gt,gm,gl,
	  ht,hm,hl,
	  lt,lm,ll,
	  fa,fb,fc,fd,
	  ga,gb,gc,gd,
	  ra,rb,rc,rd,
	  lchu,lchv,lchw,is,begTime,begBuf);

  } /* for is */

  MPI_Finalize();
}
